"""
Implementation of the Truetype file format.

Typical applications will not need to use this module directly; look at
`pyglet.font` instead.

References:
 * http://developer.apple.com/fonts/TTRefMan/RM06
 * http://www.microsoft.com/typography/otspec
"""
from __future__ import annotations

import codecs
import mmap
import os
import struct
from typing import Any


class TruetypeInfo:
    """Information about a single Truetype face.

    The class memory-maps the font file to read the tables, so
    it is vital that you call the `close` method to avoid large memory
    leaks.  Once closed, you cannot call any of the ``get_*`` methods.

    Not all tables have been implemented yet (or likely ever will).
    Currently only the name and metric tables are read; in particular
    there is no glyph or hinting information.
    """

    _name_id_lookup = {
        "copyright": 0,
        "family": 1,
        "subfamily": 2,
        "identifier": 3,
        "name": 4,
        "version": 5,
        "postscript": 6,
        "trademark": 7,
        "manufacturer": 8,
        "designer": 9,
        "description": 10,
        "vendor-url": 11,
        "designer-url": 12,
        "license": 13,
        "license-url": 14,
        "preferred-family": 16,
        "preferred-subfamily": 17,
        "compatible-name": 18,
        "sample": 19,
    }

    _platform_id_lookup = {
        "unicode": 0,
        "macintosh": 1,
        "iso": 2,
        "microsoft": 3,
        "custom": 4,
    }

    _microsoft_encoding_lookup = {
        1: "utf_16_be",
        2: "shift_jis",
        4: "big5",
        6: "johab",
        10: "utf_16_be",
    }

    _macintosh_encoding_lookup = {
        0: "mac_roman",
    }

    _character_advances: dict[str, float] | None

    def __init__(self, filename: str) -> None:
        """Read the given TrueType file.

        Args:
            filename:
                The name of any Windows, OS2 or Macintosh Truetype file.

        The object must be closed (see `close`) after use.

        An exception will be raised if the file does not exist or cannot be read.
        """
        assert filename, "must provide a font file name"
        length = os.stat(filename).st_size
        self._fileno = os.open(filename, os.O_RDONLY)
        if hasattr(mmap, "MAP_SHARED"):
            self._data = mmap.mmap(self._fileno, length, mmap.MAP_SHARED, mmap.PROT_READ)
        else:
            self._data = mmap.mmap(self._fileno, length, None, mmap.ACCESS_READ)

        self._closed = False

        offsets = _read_offset_table(self._data, 0)
        self._tables = {}
        for table in _read_table_directory_entry.array(self._data, offsets.size, offsets.num_tables):
            self._tables[table.tag] = table

        self._names = None
        self._horizontal_metrics = None
        self._character_advances = None
        self._character_kernings = None
        self._glyph_kernings = None
        self._character_map = None
        self._glyph_map = None
        self._font_selection_flags = None

        self.header = _read_head_table(self._data, self._tables["head"].offset)
        self.horizontal_header = _read_horizontal_header(self._data, self._tables["hhea"].offset)

    def get_font_selection_flags(self) -> int:
        """Return the font selection flags, as defined in OS/2 table"""
        if not self._font_selection_flags:
            OS2_table = _read_OS2_table(self._data, self._tables["OS/2"].offset)
            self._font_selection_flags = OS2_table.fs_selection
        return self._font_selection_flags

    def is_bold(self) -> bool:
        """Returns True iff the font describes itself as bold."""
        return bool(self.get_font_selection_flags() & 0x20)

    def is_italic(self) -> bool:
        """Returns True iff the font describes itself as italic."""
        return bool(self.get_font_selection_flags() & 0x1)

    def get_names(self) -> dict[tuple[int, int], tuple[int, int, str]]:
        """Returns a dictionary of names defined in the file.

        The key of each item is a tuple of ``platform_id``, ``name_id``,
        where each ID is the number as described in the Truetype format.

        The value of each item is a tuple of
        ``encoding_id``, ``language_id``, ``value``, where ``value`` is
        an encoded string.
        """
        if self._names:
            return self._names
        naming_table = _read_naming_table(self._data, self._tables["name"].offset)
        name_records = _read_name_record.array(
            self._data, self._tables["name"].offset + naming_table.size, naming_table.count)
        storage = naming_table.string_offset + self._tables["name"].offset
        self._names = {}
        for record in name_records:
            value = self._data[record.offset + storage: record.offset + storage + record.length]
            key = record.platform_id, record.name_id
            value = (record.encoding_id, record.language_id, value)
            if key not in self._names:
                self._names[key] = []
            self._names[key].append(value)
        return self._names

    def get_name(self, name: int | str, platform: int | str | None = None,
                 languages: int | str | None = None) -> str | None:
        """Returns the value of the given name in this font.

        Args:
            `name`
                Either an integer, representing the name_id desired (see
                font format); or a string describing it, see below for
                valid names.
            `platform`
                Platform for the requested name.  Can be the integer ID,
                or a string describing it.  By default, the Microsoft
                platform is searched first, then Macintosh.
            `languages`
                A list of language IDs to search.  The first language
                which defines the requested name will be used.  By default,
                all English dialects are searched.

        If the name is not found, ``None`` is returned.  If the name
        is found, the value will be decoded and returned as a unicode
        string.  Currently only some common encodings are supported.

        Valid names to request are (supply as a string)::

            'copyright'
            'family'
            'subfamily'
            'identifier'
            'name'
            'version'
            'postscript'
            'trademark'
            'manufacturer'
            'designer'
            'description'
            'vendor-url'
            'designer-url'
            'license'
            'license-url'
            'preferred-family'
            'preferred-subfamily'
            'compatible-name'
            'sample'

        Valid platforms to request are (supply as a string)::

            'unicode'
            'macintosh'
            'iso'
            'microsoft'
            'custom'
        """

        names = self.get_names()
        if isinstance(name, str):
            name = self._name_id_lookup[name]
        if not platform:
            for platform in ("microsoft", "macintosh"):
                value = self.get_name(name, platform, languages)
                if value:
                    return value
        if isinstance(platform, str):
            platform = self._platform_id_lookup[platform]
        if (platform, name) not in names:
            return None

        if platform == 3:  # setup for microsoft
            encodings = self._microsoft_encoding_lookup
            if not languages:
                # Default to english languages for microsoft
                languages = (0x409, 0x809, 0xc09, 0x1009, 0x1409, 0x1809)
        elif platform == 1:  # setup for macintosh
            encodings = self.__macintosh_encoding_lookup
            if not languages:
                # Default to english for macintosh
                languages = (0,)

        for record in names[(platform, name)]:
            if record[1] in languages and record[0] in encodings:
                decoder = codecs.getdecoder(encodings[record[0]])
                return decoder(record[2])[0]
        return None

    def get_horizontal_metrics(self) -> list:
        """Return all horizontal metric entries in table format."""
        if not self._horizontal_metrics:
            ar = _read_long_hor_metric.array(self._data,
                                             self._tables["hmtx"].offset,
                                             self.horizontal_header.number_of_h_metrics)
            self._horizontal_metrics = ar
        return self._horizontal_metrics

    def get_character_advances(self) -> dict[str, float] | None:
        """Return a dictionary of character->advance.

        They key of the dictionary is a unit-length unicode string,
        and the value is a float giving the horizontal advance in
        em.
        """
        if self._character_advances:
            return self._character_advances
        ga = self.get_glyph_advances()
        gmap = self.get_glyph_map()
        self._character_advances = {}
        for i in range(len(ga)):
            if i in gmap and gmap[i] not in self._character_advances:
                self._character_advances[gmap[i]] = ga[i]
        return self._character_advances

    def get_glyph_advances(self) -> list[float]:
        """Return a list of advances."""
        hm = self.get_horizontal_metrics()
        return [float(m.advance_width) / self.header.units_per_em for m in hm]

    def get_character_kernings(self) -> dict[tuple[str, str], int]:
        """Return a dictionary of (left,right)->kerning

        The key of the dictionary is a tuple of ``(left, right)``
        where each element is a unit-length unicode string.  The
        value of the dictionary is the horizontal pairwise kerning
        in em.
        """
        if not self._character_kernings:
            gmap = self.get_glyph_map()
            kerns = self.get_glyph_kernings()
            self._character_kernings = {}
            for pair, value in kerns.items():
                lglyph, rglyph = pair
                lchar = lglyph in gmap and gmap[lglyph] or None
                rchar = rglyph in gmap and gmap[rglyph] or None
                if lchar and rchar:
                    self._character_kernings[(lchar, rchar)] = value
        return self._character_kernings

    def get_glyph_kernings(self) -> dict[tuple[str, str], int]:
        """Return a dictionary of (left,right)->kerning

        The key of the dictionary is a tuple of ``(left, right)``
        where each element is a glyph index.  The value of the dictionary is
        the horizontal pairwise kerning in em.
        """
        if self._glyph_kernings:
            return self._glyph_kernings
        header = \
            _read_kern_header_table(self._data, self._tables["kern"].offset)
        offset = self._tables["kern"].offset + header.size
        kernings = {}
        for i in range(header.n_tables):
            header = _read_kern_subtable_header(self._data, offset)
            if header.coverage & header.horizontal_mask \
                    and not header.coverage & header.minimum_mask \
                    and not header.coverage & header.perpendicular_mask:
                if header.coverage & header.format_mask == 0:
                    self._add_kernings_format0(kernings, offset + header.size)
            offset += header.length
        self._glyph_kernings = kernings
        return kernings

    def _add_kernings_format0(self, kernings, offset):
        header = _read_kern_subtable_format0(self._data, offset)
        kerning_pairs = _read_kern_subtable_format0Pair.array(self._data,
                                                              offset + header.size, header.n_pairs)
        for pair in kerning_pairs:
            if (pair.left, pair.right) in kernings:
                kernings[(pair.left, pair.right)] += pair.value \
                                                     / float(self.header.units_per_em)
            else:
                kernings[(pair.left, pair.right)] = pair.value \
                                                    / float(self.header.units_per_em)

    def get_glyph_map(self) -> dict[int, str]:
        """Calculate and return a reverse character map.

        Returns a dictionary where the key is a glyph index and the
        value is a unit-length unicode string.
        """
        if self._glyph_map:
            return self._glyph_map
        cmap = self.get_character_map()
        self._glyph_map = {}
        for ch, glyph in cmap.items():
            if glyph not in self._glyph_map:
                self._glyph_map[glyph] = ch
        return self._glyph_map

    def get_character_map(self) -> dict[str, int]:
        """Return the character map.

        Returns a dictionary where the key is a unit-length unicode string and the value is a glyph index.  Currently
        only format 4 character maps are read.
        """
        if self._character_map:
            return self._character_map
        cmap = _read_cmap_header(self._data, self._tables["cmap"].offset)
        records = _read_cmap_encoding_record.array(self._data,
                                                   self._tables["cmap"].offset + cmap.size, cmap.num_tables)
        self._character_map = {}
        for record in records:
            if record.platform_id == 3 and record.encoding_id == 1:
                # Look at Windows Unicode charmaps only
                offset = self._tables["cmap"].offset + record.offset
                format_header = _read_cmap_format_header(self._data, offset)
                if format_header.format == 4:
                    self._character_map = \
                        self._get_character_map_format4(offset)
                    break
        return self._character_map

    def _get_character_map_format4(self, offset: int) -> dict[str, int]:
        # This is absolutely, without question, the *worst* file
        # format ever.  Whoever the fuckwit is that thought this up is
        # a fuckwit.
        header = _read_cmap_format4Header(self._data, offset)
        seg_count = header.seg_count_x2 // 2
        array_size = struct.calcsize(f">{seg_count}H")
        end_count = self._read_array(f">{seg_count}H",
                                     offset + header.size)
        start_count = self._read_array(f">{seg_count}H",
                                       offset + header.size + array_size + 2)
        id_delta = self._read_array(f">{seg_count}H",
                                    offset + header.size + array_size + 2 + array_size)
        id_range_offset_address = \
            offset + header.size + array_size + 2 + array_size + array_size
        id_range_offset = self._read_array(f">{seg_count}H",
                                           id_range_offset_address)
        character_map = {}
        for i in range(seg_count):
            if id_range_offset[i] != 0:
                if id_range_offset[i] == 65535:
                    continue  # Hack around a dodgy font (babelfish.ttf)
                for c in range(start_count[i], end_count[i] + 1):
                    addr = id_range_offset[i] + 2 * (c - start_count[i]) + \
                           id_range_offset_address + 2 * i
                    g = struct.unpack(">H", self._data[addr:addr + 2])[0]
                    if g != 0:
                        character_map[chr(c)] = (g + id_delta[i]) % 65536
            else:
                for c in range(start_count[i], end_count[i] + 1):
                    g = (c + id_delta[i]) % 65536
                    if g != 0:
                        character_map[chr(c)] = g
        return character_map

    def _read_array(self, fmt: str, offset: int) -> tuple[Any, ...]:
        size = struct.calcsize(fmt)
        return struct.unpack(fmt, self._data[offset:offset + size])

    def close(self) -> None:
        """Close the font file.

        This is a good idea, since the entire file is memory mapped in
        until this method is called.  After closing cannot rely on the
        ``get_*`` methods.
        """

        self._data.close()
        os.close(self._fileno)
        self._closed = True

    def __del__(self) -> None:
        if not self._closed:
            self.close()


def _read_table(*entries: str):
    """ Generic table constructor used for table formats listed at
     end of file."""

    fmt = ">"
    names = []
    for entry in entries:
        name, entry_type = entry.split(":")
        names.append(name)
        fmt += entry_type

    class TableClass:
        size = struct.calcsize(fmt)

        def __init__(self, data: mmap.mmap, offset: int) -> None:
            items = struct.unpack(fmt, data[offset:offset + self.size])
            self.pairs = list(zip(names, items))
            for pname, pvalue in self.pairs:
                if isinstance(pvalue, bytes):
                    pvalue = pvalue.decode("utf-8")
                setattr(self, pname, pvalue)

        def __repr__(self) -> str:
            return "{" + ", ".join([f"{pname} = {pvalue}" for pname, pvalue in self.pairs]) + "}"

        @staticmethod
        def array(data: mmap.mmap, offset: int, count: int) -> list[TableClass]:
            tables = []
            for i in range(count):
                tables.append(TableClass(data, offset))
                offset += TableClass.size
            return tables

    return TableClass


# Table formats (see references)

_read_offset_table = _read_table("scalertype:I",
                                 "num_tables:H",
                                 "search_range:H",
                                 "entry_selector:H",
                                 "range_shift:H")

_read_table_directory_entry = _read_table("tag:4s",
                                          "check_sum:I",
                                          "offset:I",
                                          "length:I")

_read_head_table = _read_table("version:i",
                               "font_revision:i",
                               "check_sum_adjustment:L",
                               "magic_number:L",
                               "flags:H",
                               "units_per_em:H",
                               "created:Q",
                               "modified:Q",
                               "x_min:h",
                               "y_min:h",
                               "x_max:h",
                               "y_max:h",
                               "mac_style:H",
                               "lowest_rec_p_pEM:H",
                               "font_direction_hint:h",
                               "index_to_loc_format:h",
                               "glyph_data_format:h")

_read_OS2_table = _read_table("version:H",
                              "x_avg_char_width:h",
                              "us_weight_class:H",
                              "us_width_class:H",
                              "fs_type:H",
                              "y_subscript_x_size:h",
                              "y_subscript_y_size:h",
                              "y_subscript_x_offset:h",
                              "y_subscript_y_offset:h",
                              "y_superscript_x_size:h",
                              "y_superscript_y_size:h",
                              "y_superscript_x_offset:h",
                              "y_superscript_y_offset:h",
                              "y_strikeout_size:h",
                              "y_strikeout_position:h",
                              "s_family_class:h",
                              "panose1:B",
                              "panose2:B",
                              "panose3:B",
                              "panose4:B",
                              "panose5:B",
                              "panose6:B",
                              "panose7:B",
                              "panose8:B",
                              "panose9:B",
                              "panose10:B",
                              "ul_unicode_range1:L",
                              "ul_unicode_range2:L",
                              "ul_unicode_range3:L",
                              "ul_unicode_range4:L",
                              "ach_vend_id:I",
                              "fs_selection:H",
                              "us_first_char_index:H",
                              "us_last_char_index:H",
                              "s_typo_ascender:h",
                              "s_typo_descender:h",
                              "s_typo_line_gap:h",
                              "us_win_ascent:H",
                              "us_win_descent:H",
                              "ul_code_page_range1:L",
                              "ul_code_page_range2:L",
                              "sx_height:h",
                              "s_cap_height:h",
                              "us_default_char:H",
                              "us_break_char:H",
                              "us_max_context:H")

_read_kern_header_table = _read_table("version_num:H",
                                      "n_tables:H")

_read_kern_subtable_header = _read_table("version:H",
                                         "length:H",
                                         "coverage:H")

_read_kern_subtable_header.horizontal_mask = 0x1
_read_kern_subtable_header.minimum_mask = 0x2
_read_kern_subtable_header.perpendicular_mask = 0x4
_read_kern_subtable_header.override_mask = 0x5
_read_kern_subtable_header.format_mask = 0xf0

_read_kern_subtable_format0 = _read_table("n_pairs:H",
                                          "search_range:H",
                                          "entry_selector:H",
                                          "range_shift:H")
_read_kern_subtable_format0Pair = _read_table("left:H",
                                              "right:H",
                                              "value:h")

_read_cmap_header = _read_table("version:H",
                                "num_tables:H")

_read_cmap_encoding_record = _read_table("platform_id:H",
                                         "encoding_id:H",
                                         "offset:L")

_read_cmap_format_header = _read_table("format:H",
                                       "length:H")
_read_cmap_format4Header = _read_table("format:H",
                                       "length:H",
                                       "language:H",
                                       "seg_count_x2:H",
                                       "search_range:H",
                                       "entry_selector:H",
                                       "range_shift:H")

_read_horizontal_header = _read_table("version:i",
                                      "Advance:h",
                                      "Descender:h",
                                      "LineGap:h",
                                      "advance_width_max:H",
                                      "min_left_side_bearing:h",
                                      "min_right_side_bearing:h",
                                      "x_max_extent:h",
                                      "caret_slope_rise:h",
                                      "caret_slope_run:h",
                                      "caret_offset:h",
                                      "reserved1:h",
                                      "reserved2:h",
                                      "reserved3:h",
                                      "reserved4:h",
                                      "metric_data_format:h",
                                      "number_of_h_metrics:H")

_read_long_hor_metric = _read_table("advance_width:H",
                                    "lsb:h")

_read_naming_table = _read_table("format:H",
                                 "count:H",
                                 "string_offset:H")

_read_name_record = _read_table("platform_id:H",
                                "encoding_id:H",
                                "language_id:H",
                                "name_id:H",
                                "length:H",
                                "offset:H")



"""
Tests for events on windows.
"""
import pytest
import random

from pyglet import font
from pyglet import gl
from pyglet.window import key, Window
from pyglet.window.event import WindowEventLogger

from tests.base.interactive import InteractiveTestCase
from tests.interactive.window import window_util


class WindowEventsTestCase(InteractiveTestCase):
    """
    Base class that shows a window displaying instructions for the test. Then it waits for events
    to continue. Optionally the user can fail the test by pressing Escape.
    """

    # Defaults
    window_size = 400, 200
    window = None
    question = None

    def setUp(self):
        self.finished = False
        self.failure = None
        self.label = None

    def fail_test(self, failure):
        self.failure = failure
        self.finished = True

    def pass_test(self):
        self.finished = True

    def _render_question(self):
        fnt = font.load('Courier')
        self.label = font.Text(fnt, text=self.question, x=10, y=self.window_size[1]-20)

    def _draw(self):
        gl.glClearColor(0.5, 0, 0, 1)
        gl.glClear(gl.GL_COLOR_BUFFER_BIT)
        gl.glLoadIdentity()
        self.label.draw()
        self.window.flip()

    def _test_main(self):
        assert self.question

        width, height = self.window_size
        self.window = w = Window(width, height, visible=False, resizable=False)
        try:
            w.push_handlers(self)
            self._render_question()
            w.set_visible()

            while not self.finished and not w.has_exit:
                self._draw()
                w.dispatch_events()

        finally:
            w.close()

        # TODO: Allow entering reason of failure if user aborts
        self.assertTrue(self.finished, msg="Test aborted")
        self.assertIsNone(self.failure, msg=self.failure)


@pytest.mark.requires_user_action
class KeyPressWindowEventTestCase(WindowEventsTestCase):
    number_of_checks = 10
    keys = (key.A, key.B, key.C, key.D, key.E, key.F, key.G, key.H, key.I, key.J, key.K, key.L,
            key.M, key.N, key.O, key.P, key.Q, key.R, key.S, key.T, key.U, key.V, key.W, key.X,
            key.Y, key.Z)
    mod_shift_keys = (key.LSHIFT, key.RSHIFT)
    mod_ctrl_keys = (key.LCTRL, key.RCTRL)
    mod_alt_keys = (key.LALT, key.RALT)
    mod_option_keys = (key.LOPTION, key.ROPTION)
    mod_meta_keys = (key.LMETA, key.RMETA)
    mod_meta = key.MOD_SHIFT | key.MOD_ALT

    def setUp(self):
        super().setUp()
        self.chosen_symbol = None
        self.chosen_modifiers = None
        self.completely_pressed = False
        self.active_keys = []
        self.checks_passed = 0

    def on_key_press(self, symbol, modifiers):
        print('Press: ', key.symbol_string(symbol))
        self.active_keys.append(symbol)

        if self.completely_pressed:
            self.fail_test('Key already pressed, no release received.')

        elif self._is_correct_modifier_key(symbol):
            # Does not seem to be correct for modifier keys
            #self._check_modifiers_against_pressed_keys(modifiers)
            pass

        elif self._is_correct_key(symbol):
            self._check_modifiers_against_pressed_keys(modifiers)
            self.completely_pressed = True

    def on_key_release(self, symbol, modifiers):
        print('Release: ', key.symbol_string(symbol))
        symbol = self._handle_meta_release(symbol)
        if symbol not in self.active_keys:
            self.fail_test('Released key "{}" was not pressed before.'.format(key.symbol_string(symbol)))
        else:
            self.active_keys.remove(symbol)

        if len(self.active_keys) == 0 and self.completely_pressed:
            self.completely_pressed = False
            self.checks_passed += 1
            if self.checks_passed == self.number_of_checks:
                self.pass_test()
            else:
                self._select_next_key()

    def _select_next_key(self):
        self.chosen_symbol = random.choice(self.keys)

        # Little trick, Ctrl, Alt and Shift are lowest modifier values, so everything between 0 and
        # the full combination is a permutation of these three.
        max_modifiers = key.MOD_SHIFT | key.MOD_ALT | key.MOD_CTRL
        # Give a little more weight to key without modifiers
        self.chosen_modifiers = max(0, random.randint(-2, max_modifiers))

        self._update_question()

    def _update_question(self):
        modifiers = []
        if self.chosen_modifiers & key.MOD_SHIFT:
            modifiers.append('<Shift>')
        if self.chosen_modifiers & key.MOD_ALT:
            modifiers.append('<Alt/Option>')
        if self.chosen_modifiers & key.MOD_CTRL:
            modifiers.append('<Ctrl>')

        self.question = """Please press and release the following combination of keys.
Only use <Shift> if explicitly asked to do so.

{} {}


Press Esc if test does not pass.""".format(' '.join(modifiers), key.symbol_string(self.chosen_symbol))
        self._render_question()

    def _is_correct_modifier_key(self, symbol):
        modifier = self._get_modifier_for_key(symbol)
        if modifier == 0:
            return False

        if modifier == key.MOD_OPTION:
            # Ignore option key as it doubles with Alt
            return True

        if not self.chosen_modifiers & modifier:
            self.fail_test('Unexpected modifier key "{}"'.format(key.symbol_string(symbol)))

        return True

    def _get_modifier_for_key(self, symbol):
        if symbol in self.mod_shift_keys:
            return key.MOD_SHIFT
        elif symbol in self.mod_alt_keys:
            return key.MOD_ALT
        elif symbol in self.mod_ctrl_keys:
            return key.MOD_CTRL
        elif symbol in self.mod_option_keys:
            return key.MOD_OPTION
        elif symbol in self.mod_meta_keys:
            return self.mod_meta
        else:
            return 0

    def _get_modifiers_from_pressed_keys(self):
        modifiers = 0
        for symbol in self.active_keys:
            modifiers |= self._get_modifier_for_key(symbol)
        return modifiers

    def _translate_option_modifier(self, modifiers):
        """
        On macOS the Alt key is also called the Option key. Key presses for both Alt and Option
        are emitted, but the modifier list contains only one of them. Unify modifiers to
        contain only Alt if Alt and/or Option is active.
        """
        if modifiers & key.MOD_OPTION or modifiers & key.MOD_ALT:
            return (modifiers & ~key.MOD_OPTION) | key.MOD_ALT
        else:
            return modifiers

    def _check_modifiers_against_pressed_keys(self, modifiers):
        modifiers_from_keys = self._translate_option_modifier(self._get_modifiers_from_pressed_keys())
        modifiers = self._translate_option_modifier(modifiers)
        if modifiers != modifiers_from_keys:
            self.fail_test('Received modifiers "{}" do not match pressed keys "{}"'.format(
                                key.modifiers_string(modifiers),
                                key.modifiers_string(modifiers_from_keys)))

    def _is_correct_key(self, symbol):
        if not self.chosen_symbol:
            self.fail_test('No more key presses/releases expected.')
            return False

        if self.chosen_symbol != symbol:
            self.fail_test('Received key "{}", but expected "{}"'.format(key.symbol_string(symbol),
                                                                         key.symbol_string(self.chosen_symbol)))
            return False

        return True

    def _handle_meta_release(self, symbol):
        """The meta key can be either released as meta or as alt shift or vv"""
        if symbol in (key.LMETA, key.LALT, key.RMETA, key.RALT):
            if symbol not in self.active_keys:
                if symbol == key.LMETA and key.LALT in self.active_keys:
                    return key.LALT
                if symbol == key.RMETA and key.RALT in self.active_keys:
                    return key.RALT
                if symbol == key.LALT and key.LMETA in self.active_keys:
                    return key.LMETA
                if symbol == key.RALT and key.RMETA in self.active_keys:
                    return key.RMETA
        return symbol

    def test_key_press_release(self):
        """Show several keys to press. Check that the event is triggered for the correct key."""
        self._select_next_key()
        self._test_main()


@pytest.mark.requires_user_action
class TextWindowEventsTest(WindowEventsTestCase):
    number_of_checks = 10
    text = '`1234567890-=~!@#$%^&*()_+qwertyuiop[]\\QWERTYUIOP{}|asdfghjkl;\'ASDFGHJKL:"zxcvbnm,./ZXCVBNM<>?'

    def setUp(self):
        super().setUp()
        self.chosen_text = None
        self.checks_passed = 0

    def on_text(self, text):
        if text != self.chosen_text:
            self.fail_test('Expected "{}", received "{}"'.format(self.chosen_text, text))
        else:
            self.checks_passed += 1
            if self.checks_passed >= self.number_of_checks:
                self.pass_test()
            else:
                self._select_next_text()

    def _select_next_text(self):
        self.chosen_text = random.choice(self.text)
        self._update_question()

    def _update_question(self):
        self.question = """Please type the following character exactly.
Use <Shift> if needed.

{}


Press Esc if test does not pass.""".format(self.chosen_text)
        self._render_question()

    def test_key_text(self):
        """Show several keys to press. Check that the text events are triggered correctly."""
        self._select_next_text()
        self._test_main()


@pytest.mark.requires_user_action
class TextMotionWindowEventsTest(WindowEventsTestCase):
    number_of_checks = 10
    motion_keys = (key.MOTION_UP, key.MOTION_RIGHT, key.MOTION_DOWN, key.MOTION_LEFT,
                   key.MOTION_NEXT_PAGE, key.MOTION_PREVIOUS_PAGE, key.MOTION_BACKSPACE,
                   key.MOTION_DELETE)

    def setUp(self):
        super().setUp()
        self.chosen_key = None
        self.checks_passed = 0

    def on_key_press(self, symbol, modifiers):
        if symbol == key.X:
            self._select_next_key()

    def on_text_motion(self, motion):
        if motion != self.chosen_key:
            self.fail_test('Expected "{}", received "{}"'.format(
                key.motion_string(self.chosen_key), key.motion_string(motion)))
        else:
            self.checks_passed += 1
            if self.checks_passed >= self.number_of_checks:
                self.pass_test()
            else:
                self._select_next_key()

    def _select_next_key(self):
        self.chosen_key = random.choice(self.motion_keys)
        self._update_question()

    def _update_question(self):
        self.question = """Please press:

{} ({})


Press the X key if you do not have this motion key.
Press Esc if test does not pass.""".format(key.motion_string(self.chosen_key),
                                           key.symbol_string(self.chosen_key))
        self._render_question()

    def test_key_text_motion(self):
        """Show several motion keys to press. Check that the on_text_motion events are triggered
        correctly."""
        self._select_next_key()
        self._test_main()

@pytest.mark.requires_user_action
class TextMotionSelectWindowEventsTest(WindowEventsTestCase):
    number_of_checks = 10
    motion_keys = (key.MOTION_UP, key.MOTION_RIGHT, key.MOTION_DOWN, key.MOTION_LEFT,
                   key.MOTION_NEXT_PAGE, key.MOTION_PREVIOUS_PAGE, key.MOTION_BACKSPACE,
                   key.MOTION_DELETE)

    def setUp(self):
        super().setUp()
        self.chosen_key = None
        self.checks_passed = 0

    def on_key_press(self, symbol, modifiers):
        if symbol == key.X:
            self._select_next_key()

    def on_text_motion_select(self, motion):
        if motion != self.chosen_key:
            self.fail_test('Expected "{}", received "{}"'.format(
                key.motion_string(self.chosen_key), key.motion_string(motion)))
        else:
            self.checks_passed += 1
            if self.checks_passed >= self.number_of_checks:
                self.pass_test()
            else:
                self._select_next_key()

    def _select_next_key(self):
        self.chosen_key = random.choice(self.motion_keys)
        self._update_question()

    def _update_question(self):
        self.question = """Please hold <Shift> and press:

{} ({})


Press the X key if you do not have this motion key.
Press Esc if test does not pass.""".format(key.motion_string(self.chosen_key),
                                           key.symbol_string(self.chosen_key))
        self._render_question()

    def test_key_text_motion_select(self):
        """Show several motion keys to press. Check that the on_text_motion_select events are
        triggered correctly combined with shift."""
        self._select_next_key()
        self._test_main()


@pytest.mark.requires_user_action
class CloseWindowEventsTest(WindowEventsTestCase):
    def on_close(self):
        self.pass_test()

    def test_on_close_event(self):
        """Test the on_close event triggered when closing the window."""
        self.question = "Please close this window by\nclicking the close button."
        self._test_main()


@pytest.mark.requires_user_action
class ActivateDeactivateWindowEventsTest(WindowEventsTestCase):
    number_of_checks = 3

    def setUp(self):
        super().setUp()
        self.window_active = None
        self.checks_passed = 0

    def on_expose(self):
        self.window_active = True
        self._update_question()

    def on_activate(self):
        if self.window_active:
            self.fail_test('Got double on_activate')
        else:
            self.window_active = True
            self.checks_passed += 1
            if self.checks_passed >= self.number_of_checks:
                self.pass_test()
            else:
                self._update_question()

    def on_deactivate(self):
        if not self.window_active:
            self.fail_test('Got double on_deactivate')
        else:
            self.window_active = False
            self._update_question()

    def _update_question(self):
        if self.window_active:
            self.question = "Please activate another window."
        else:
            self.question = "Please activate this window."
        self._render_question()

    def test_activate_deactivate(self):
        """Test the on_activate and on_deactivate events triggered when the window gets activated
        or deactivated."""
        self._update_question()
        self._test_main()


@pytest.mark.requires_user_action
class ExposeWindowEventsTest(WindowEventsTestCase):
    number_of_checks = 5

    def setUp(self):
        super().setUp()
        self.checks_passed = 0

    def on_expose(self):
        self.checks_passed += 1
        if self.checks_passed >= self.number_of_checks:
            self.pass_test()

    def test_expose(self):
        """Test the on_expose event triggered when a redraw of the window is required."""
        self.question = ("Please trigger a redraw of this window.\n\n"
                         "Depending on your OS and window manager you might need to:\n"
                         "- Cover the window with another window and uncover again\n"
                         "- Minimize and restore the window\n\n"
                         "Repeat up to 5 times (less might be accepted due to initial drawing)")
        self.window_size = 700, 200
        self._test_main()


@pytest.mark.requires_user_action
class ShowHideWindowEventsTest(WindowEventsTestCase):
    number_of_checks = 5

    def setUp(self):
        super().setUp()
        self.checks_passed = 0
        self.visible = False

    def on_show(self):
        if self.visible:
            self.fail_test('Received on_show twice without on_hide')
        else:
            self.checks_passed += 1
            self.visible = True
            if self.checks_passed >= self.number_of_checks:
                self.pass_test()

    def on_hide(self):
        if not self.visible:
            self.fail_test('Received on_hide twice without on_show')
        else:
            self.visible = False

    def test_show_hide(self):
        """Test the on_show and on_hide events."""
        self.question = ('Please trigger hide and show this window again.\n'
                         'You can do this by:\n'
                         '- Minimize and restore the window\n'
                         '- On OS X show and hide using Command+H or the dock context menu\n'
                         '\n'
                         'Test passes after doing this 4 times.')
        self.window_size = 700, 200
        self._test_main()


@pytest.mark.requires_user_action
class EVENT_BUTTON(InteractiveTestCase):
    """Test that mouse button events work correctly.

    Expected behaviour:
        One window will be opened.  Click within this window and check the console
        output for mouse events.  
        - Buttons 1, 2, 4 correspond to left, middle, right, respectively.
        - No events for scroll wheel
        - Modifiers are correct

        Close the window or press ESC to end the test.
    """
    def on_mouse_press(self, x, y, button, modifiers):
        print('Mouse button %d pressed at %f,%f with %s' % \
            (button, x, y, key.modifiers_string(modifiers)))

    def on_mouse_release(self, x, y, button, modifiers):
        print('Mouse button %d released at %f,%f with %s' % \
            (button, x, y, key.modifiers_string(modifiers)))

    def test_button(self):
        w = Window(200, 200)
        try:
            w.push_handlers(self)
            while not w.has_exit:
                w.dispatch_events()
        finally:
            w.close()
        self.user_verify('Pass test?', take_screenshot=False)


@pytest.mark.requires_user_action
class EVENT_MOVE(InteractiveTestCase):
    """Test that window move event works correctly.

    Expected behaviour:
        One window will be opened.  Move the window and ensure that the
        location printed to the terminal is correct.

        Close the window or press ESC to end the test.
    """
    def on_move(self, x, y):
        print('Window moved to %dx%d.' % (x, y))

    def test_move(self):
        w = Window(200, 200)
        try:
            w.push_handlers(self)
            while not w.has_exit:
                w.dispatch_events()
        finally:
            w.close()
        self.user_verify('Pass test?', take_screenshot=False)


@pytest.mark.requires_user_action
class EVENT_RESIZE(InteractiveTestCase):
    """Test that resize event works correctly.

    Expected behaviour:
        One window will be opened.  Resize the window and ensure that the
        dimensions printed to the terminal are correct.  You should see
        a green border inside the window but no red.

        Close the window or press ESC to end the test.
    """
    def on_resize(self, width, height):
        print('Window resized to %dx%d.' % (width, height))

    def test_resize(self):
        w = Window(200, 200, resizable=True)
        try:
            w.push_handlers(self)
            while not w.has_exit:
                window_util.draw_client_border(w)
                w.flip()
                w.dispatch_events()
        finally:
            w.close()
        self.user_verify('Pass test?', take_screenshot=False)


@pytest.mark.requires_user_action
class EVENT_MOUSE_DRAG(InteractiveTestCase):
    """Test that mouse drag event works correctly.

    Expected behaviour:
        One window will be opened.  Click and drag with the mouse and ensure
        that buttons, coordinates and modifiers are reported correctly.  Events
        should be generated even when the drag leaves the window.

        Close the window or press ESC to end the test.
    """
    def test_mouse_drag(self):
        w = Window(200, 200)
        try:
            w.push_handlers(WindowEventLogger())
            while not w.has_exit:
                w.dispatch_events()
        finally:
            w.close()
        self.user_verify('Pass test?', take_screenshot=False)


@pytest.mark.requires_user_action
class EVENT_MOUSEMOTION(InteractiveTestCase):
    """Test that mouse motion event works correctly.

    Expected behaviour:
        One window will be opened.  Move the mouse in and out of this window
        and ensure the absolute and relative coordinates are correct.
        - Absolute coordinates should have (0,0) at bottom-left of client area
        of window with positive y-axis pointing up and positive x-axis right.
        - Relative coordinates should be positive when moving up and right.

        Close the window or press ESC to end the test.
    """
    def on_mouse_motion(self, x, y, dx, dy):
        print('Mouse at (%f, %f); relative (%f, %f).' % \
            (x, y, dx, dy))

    def test_motion(self):
        w = Window(200, 200)
        try:
            w.push_handlers(self)
            while not w.has_exit:
                w.dispatch_events()
        finally:
            w.close()
        self.user_verify('Pass test?', take_screenshot=False)


@pytest.mark.requires_user_action
class EVENT_MOUSEMOTION_EXCLUSIVE(InteractiveTestCase):
    """Test that mouse motion event works correctly in exclusive mode.

    Expected behaviour:
        One window will be opened. The mouse cursor will disappear. Move the 
        mouse and ensure the relative coordinates are correct.
        - Absolute coordinates should be 0, 0.
        - Relative coordinates should be positive when moving up and right.

        Try also to use the mouse buttons and mouse wheel.
        - Buttons 1, 2, 4 correspond to left, middle, right, respectively.
        - Buttons press and release will display the button and key modifiers.
        - Drag motions show relative movement, buttons pressed and key 
          modifier.
        - Mouse wheel show positive value for forward movements.

        Close the window or press ESC to end the test.
    """
    def on_mouse_motion(self, x, y, dx, dy):
        print('Mouse at (%f, %f); relative (%f, %f).' % \
            (x, y, dx, dy))

    def on_mouse_press(self, x, y, button, modifiers):
        print('Mouse button %d pressed at %f,%f with %s' % \
            (button, x, y, key.modifiers_string(modifiers)))

    def on_mouse_release(self, x, y, button, modifiers):
        print('Mouse button %d released at %f,%f with %s' % \
            (button, x, y, key.modifiers_string(modifiers)))

    def on_mouse_drag(self, x, y, dx, dy, button, modifiers):
        print('Mouse dragged. Button %d pressed. dx, dy %f,%f with %s' % \
            (button, dx, dy, key.modifiers_string(modifiers)))

    def on_mouse_scroll(self, x, y, scrollx, scrolly):
        print('Mouse scroll %f' % (scrolly))


    def test_motion(self):
        w = Window(200, 200)
        try:
            w.set_exclusive_mouse(True)
            w.push_handlers(self)
            while not w.has_exit:
                w.dispatch_events()
        finally:
            w.set_exclusive_mouse(False)
            w.close()

        self.user_verify('Pass test?', take_screenshot=False)


@pytest.mark.requires_user_action
class EVENT_MOUSE_SCROLL(InteractiveTestCase):
    """Test that mouse scroll event works correctly.

    Expected behaviour:
        One window will be opened.  Move the scroll wheel and check that events
        are printed to console.  Positive values are associated with scrolling
        up.

        Scrolling can also be side-to-side, for example with an Apple Mighty
        Mouse.

        The actual scroll value is dependent on your operating system
        user preferences.

        Close the window or press ESC to end the test.
    """
    def on_mouse_scroll(self, x, y, dx, dy):
        print('Mouse scrolled (%f, %f) (x=%f, y=%f)' % (dx, dy, x, y))

    def test_mouse_scroll(self):
        w = Window(200, 200)
        try:
            w.push_handlers(self)
            while not w.has_exit:
                w.dispatch_events()
        finally:
            w.close()
        self.user_verify('Pass test?', take_screenshot=False)


@pytest.mark.requires_user_action
class EVENT_MOUSE_ENTER_LEAVE(InteractiveTestCase):
    """Test that mouse enter and leave events work correctly.

    Expected behaviour:
        One window will be opened.  Move the mouse in and out of this window
        and ensure the events displayed are correct.

        Close the window or press ESC to end the test.
    """
    def on_mouse_enter(self, x, y):
        print('Entered at %f, %f' % (x, y))

    def on_mouse_leave(self, x, y):
        print('Left at %f, %f' % (x, y))

    def test_motion(self):
        w = Window(200, 200)
        try:
            w.push_handlers(self)
            while not w.has_exit:
                w.dispatch_events()
        finally:
            w.close()
        self.user_verify('Pass test?', take_screenshot=False)



"""Test creation of all Label classes and decoders"""
import random

import pytest


from pyglet.text import decode_text, decode_attributed, decode_html
from pyglet.text import Label, DocumentLabel, HTMLLabel

WIDTH = 500
HEIGHT = 100
X = random.randint(0, 900)
Y = random.randint(0, 600)
Z = random.randint(-10, 10)


@pytest.mark.parametrize('label_class', [Label, HTMLLabel])
def test_label_creation(label_class):
    label = label_class("This is a test", x=X, y=Y, z=Z)
    assert label.x == X
    assert label.y == Y
    assert label.z == Z


@pytest.fixture(params=[(decode_text, "This is a string of regular text."),
                        (decode_html, "<font color=green>This is html text.</font>"),
                        (decode_attributed, "This is {bold True}attributed{bold False} text.")])
def document(request):
    decoder, string = request.param
    return decoder(string)


def test_documentlabel_creation(document):
    label = DocumentLabel(document=document, x=X, y=Y, z=Z)
    assert label.x == X
    assert label.y == Y
    assert label.z == Z



import random
from . import physicalobject, resources


class Asteroid(physicalobject.PhysicalObject):
    """An asteroid that divides a little before it dies"""

    def __init__(self, *args, **kwargs):
        super().__init__(resources.asteroid_image, *args, **kwargs)

        # Slowly rotate the asteroid as it moves
        self.rotate_speed = random.random() * 100.0 - 50.0

    def update(self, dt):
        super().update(dt)
        self.rotation += self.rotate_speed * dt

    def handle_collision_with(self, other_object):
        super().handle_collision_with(other_object)

        # Superclass handles deadness already
        if self.dead and self.scale > 0.25:
            num_asteroids = random.randint(2, 3)
            for i in range(num_asteroids):
                new_asteroid = Asteroid(x=self.x, y=self.y, batch=self.batch)
                new_asteroid.rotation = random.randint(0, 360)
                new_asteroid.velocity_x = random.random() * 70 + self.velocity_x
                new_asteroid.velocity_y = random.random() * 70 + self.velocity_y
                new_asteroid.scale = self.scale * 0.5
                self.new_objects.append(new_asteroid)



#!/usr/bin/env python
"""Demonstrates how to manage two independent windows.
"""

import pyglet


w1 = pyglet.window.Window(200, 200, caption='First window', resizable=True)
w1.switch_to()
s1 = pyglet.shapes.Rectangle(100, 100, 100, 100, color=(50, 50, 200))
s1.anchor_position = 50, 50


@w1.event
def on_draw():
    w1.clear()
    s1.rotation -= 0.5
    s1.draw()


w2 = pyglet.window.Window(300, 300, caption='Second window', resizable=True)
w2.switch_to()
s2 = pyglet.shapes.Rectangle(150, 150, 150, 150, color=(50, 200, 50))
s2.anchor_position = 75, 75


@w2.event
def on_draw():
    w2.clear()
    s2.rotation += 0.5
    s2.draw()


pyglet.app.run()



"""Use ffmpeg to decode audio and video media."""
from __future__ import annotations

import sys
from collections import deque
from ctypes import (
    POINTER,
    Array,
    Structure,
    addressof,
    byref,
    c_char_p,
    c_int,
    c_int32,
    c_uint8,
    cast,
    create_string_buffer,
    memmove,
)
from dataclasses import dataclass
from typing import BinaryIO, TYPE_CHECKING, Iterator, Sequence

from _ctypes import _Pointer

import pyglet
import pyglet.lib
from pyglet import image
from pyglet.util import asbytes, asstr
from . import MediaDecoder
from .base import AudioData, AudioFormat, SourceInfo, StaticSource, StreamingSource, VideoFormat
from .ffmpeg_lib import (
    AV_CODEC_ID_VP8,
    AV_CODEC_ID_VP9,
    FF_INPUT_BUFFER_PADDING_SIZE,
    SWS_FAST_BILINEAR,
    AVPacket,
    SwrContext,
    SwsContext,
    avcodec,
    libavformat,
    libavutil,
    swscale,
)
from .ffmpeg_lib.libavformat import AVCodecContext, AVFormatContext, avformat, avformat_version
from .ffmpeg_lib.libavutil import (
    AV_NOPTS_VALUE,
    AV_PIX_FMT_RGBA,
    AV_SAMPLE_FMT_FLT,
    AV_SAMPLE_FMT_FLTP,
    AV_SAMPLE_FMT_S16,
    AV_SAMPLE_FMT_S16P,
    AV_SAMPLE_FMT_S32,
    AV_SAMPLE_FMT_S32P,
    AV_SAMPLE_FMT_U8,
    AV_SAMPLE_FMT_U8P,
    AV_TIME_BASE,
    AV_TIME_BASE_Q,
    AVMEDIA_TYPE_AUDIO,
    AVMEDIA_TYPE_VIDEO,
    AVChannelLayout,
    AVFrame,
    AVRational,
    avutil,
)
from .ffmpeg_lib.libswresample import swresample, swresample_version
from ..exceptions import MediaFormatException

if TYPE_CHECKING:
    from .ffmpeg_lib.libavformat import AVStream


class FileInfo:
    def __init__(self):
        self.n_streams = None
        self.start_time = None
        self.duration = None
        self.title = ""
        self.author = ""
        self.copyright = ""
        self.comment = ""
        self.album = ""
        self.year = None
        self.track = ""
        self.genre = ""


@dataclass
class StreamVideoInfo:
    width: int
    height: int
    sample_aspect_num: int
    sample_aspect_den: int
    frame_rate_num: int
    frame_rate_den: int
    codec_id: int


@dataclass
class StreamAudioInfo:
    sample_format: int
    sample_rate: int
    channels: int
    sample_bits: int | None = None


class FFmpegFile(Structure):
    _fields_ = [
        ('context', POINTER(AVFormatContext)),
    ]


class FFmpegStream(Structure):
    _fields_ = [
        ('type', c_int32),
        ('format_context', POINTER(AVFormatContext)),
        ('codec_context', POINTER(AVCodecContext)),
        ('frame', POINTER(AVFrame)),
        ('time_base', AVRational),
    ]


class FFmpegException(MediaFormatException):
    pass


def ffmpeg_get_audio_buffer_size(audio_format):
    """Return the audio buffer size

    Buffer size can accommodate 1 sec of audio data.
    """
    return audio_format.bytes_per_second + FF_INPUT_BUFFER_PADDING_SIZE


def ffmpeg_init():
    """Initialize libavformat and register all the muxers, demuxers and protocols."""


class MemoryFileObject:
    """A class to manage reading and seeking of a ffmpeg file object."""
    buffer_size = 32768

    def __init__(self, file: BinaryIO) -> None:
        self.file = file
        self.fmt_context = None
        self.buffer = None

        if not getattr(self.file, 'seek', None) or not getattr(self.file, 'tell', None):
            raise Exception("File object does not support seeking.")

        # Seek to end of file to get the filesize.
        self.file.seek(0, 2)
        self.file_size = self.file.tell()
        self.file.seek(0)  # Put cursor back at the beginning.

        def read_data_cb(_, buff: bytes, buf_size: int) -> int:
            data = self.file.read(buf_size)
            read_size = len(data)
            memmove(buff, data, read_size)
            return read_size

        def seek_data_cb(_, offset: int, whence: int) -> int:
            if whence == libavformat.AVSEEK_SIZE:
                return self.file_size

            pos = self.file.seek(offset, whence)
            return pos

        self.read_func = libavformat.ffmpeg_read_func(read_data_cb)
        self.seek_func = libavformat.ffmpeg_seek_func(seek_data_cb)

    def __del__(self) -> None:
        """These are usually freed when the source is, but no guarantee."""
        if self.buffer:
            try:
                avutil.av_freep(self.buffer)
            except OSError:
                pass

        if self.fmt_context:
            try:
                avutil.av_freep(self.fmt_context)
            except OSError:
                pass


def ffmpeg_open_memory_file(filename: bytes, file_object: BinaryIO) -> tuple[FFmpegFile, MemoryFileObject]:
    """Open a media file from a file-like object.

    Returns:
        The structure containing all the information for the media and the memory file object.
    """
    file = FFmpegFile()

    file.context = libavformat.avformat.avformat_alloc_context()
    file.context.contents.seekable = 1

    memory_file = MemoryFileObject(file_object)

    av_buf = libavutil.avutil.av_malloc(memory_file.buffer_size)
    memory_file.buffer = cast(av_buf, c_char_p)

    ptr = create_string_buffer(memory_file.buffer_size)

    memory_file.fmt_context = libavformat.avformat.avio_alloc_context(
        memory_file.buffer,
        memory_file.buffer_size,
        0,
        ptr,
        memory_file.read_func,
        None,
        memory_file.seek_func,
    )

    file.context.contents.pb = memory_file.fmt_context
    file.context.contents.flags |= libavformat.AVFMT_FLAG_CUSTOM_IO

    result = avformat.avformat_open_input(byref(file.context), filename, None, None)

    if result != 0:
        raise FFmpegException('avformat_open_input in ffmpeg_open_filename returned an error opening file '
                              + filename.decode("utf8")
                              + ' Error code: ' + str(result))

    result = avformat.avformat_find_stream_info(file.context, None)
    if result < 0:
        raise FFmpegException('Could not find stream info')

    return file, memory_file


def ffmpeg_open_filename(filename: bytes) -> FFmpegFile:
    """Open the media file.

    Returns:
        The structure containing all the information for the media.
    """
    file = FFmpegFile()  # TODO: delete this structure and use directly AVFormatContext
    result = avformat.avformat_open_input(byref(file.context),
                                          filename,
                                          None,
                                          None)

    if result != 0:
        raise FFmpegException('avformat_open_input in ffmpeg_open_filename returned an error opening file '
                              + filename.decode("utf8")
                              + ' Error code: ' + str(result))

    result = avformat.avformat_find_stream_info(file.context, None)
    if result < 0:
        raise FFmpegException('Could not find stream info')

    return file


def ffmpeg_close_file(file: FFmpegFile):
    """Close the media file and free resources."""
    avformat.avformat_close_input(byref(file.context))


def ffmpeg_file_info(file: FFmpegFile) -> FileInfo:
    """Get information on the file:

        - number of streams
        - duration
        - artist
        - album
        - date
        - track

    Returns:
        The file info instance containing all the meta information.
    """
    info = FileInfo()
    info.n_streams = file.context.contents.nb_streams
    info.start_time = file.context.contents.start_time
    info.duration = file.context.contents.duration

    entry = avutil.av_dict_get(file.context.contents.metadata, asbytes('title'), None, 0)
    if entry:
        info.title = asstr(entry.contents.value)

    entry = avutil.av_dict_get(file.context.contents.metadata, asbytes('artist'), None, 0) \
        or avutil.av_dict_get(file.context.contents.metadata, asbytes('album_artist'), None, 0)
    if entry:
        info.author = asstr(entry.contents.value)

    entry = avutil.av_dict_get(file.context.contents.metadata, asbytes('copyright'), None, 0)
    if entry:
        info.copyright = asstr(entry.contents.value)

    entry = avutil.av_dict_get(file.context.contents.metadata, asbytes('comment'), None, 0)
    if entry:
        info.comment = asstr(entry.contents.value)

    entry = avutil.av_dict_get(file.context.contents.metadata, asbytes('album'), None, 0)
    if entry:
        info.album = asstr(entry.contents.value)

    entry = avutil.av_dict_get(file.context.contents.metadata, asbytes('date'), None, 0)
    if entry:
        info.year = asstr(entry.contents.value)

    entry = avutil.av_dict_get(file.context.contents.metadata, asbytes('track'), None, 0)
    if entry:
        info.track = asstr(entry.contents.value)

    entry = avutil.av_dict_get(file.context.contents.metadata, asbytes('genre'), None, 0)
    if entry:
        info.genre = asstr(entry.contents.value)

    return info


def ffmpeg_stream_info(file: FFmpegFile, stream_index: int) -> StreamAudioInfo | StreamVideoInfo | None:
    """Open the stream and retrieve information about it."""
    av_stream = file.context.contents.streams[stream_index].contents

    context = av_stream.codecpar.contents

    if context.codec_type == AVMEDIA_TYPE_VIDEO:
        if _debug:
            print(" codec_type=", context.codec_type)
            print(" codec_id=", context.codec_id)
            print(" codec name=", avcodec.avcodec_get_name(context.codec_id).decode('utf-8'))
            print(" codec_tag=", context.codec_tag)
            print(" extradata=", context.extradata)
            print(" extradata_size=", context.extradata_size)
            print(" format=", context.format)
            print(" bit_rate=", context.bit_rate)
            print(" bits_per_coded_sample=", context.bits_per_coded_sample)
            print(" bits_per_raw_sample=", context.bits_per_raw_sample)
            print(" profile=", context.profile)
            print(" level=", context.level)
            print(" width=", context.width)
            print(" height=", context.height)
            print(" sample_aspect_ratio=", context.sample_aspect_ratio.num, context.sample_aspect_ratio.den)
            print(" field_order=", context.field_order)
            print(" color_range=", context.color_range)
            print(" color_primaries=", context.color_primaries)
            print(" color_trc=", context.color_trc)
            print(" color_space=", context.color_space)
            print(" chroma_location=", context.chroma_location)
            print(" video_delay=", context.video_delay)
            if avformat_version >= 61:
                print(" channel_layout=", context.ch_layout)
            else:
                print(" channel_layout=", context.channel_layout)
                print(" channels=", context.channels)
            print(" sample_rate=", context.sample_rate)
            print(" block_align=", context.block_align)
            print(" frame_size=", context.frame_size)
            print(" initial_padding=", context.initial_padding)
            print(" trailing_padding=", context.trailing_padding)
            print(" seek_preroll=", context.seek_preroll)
        #
        frame_rate = avformat.av_guess_frame_rate(file.context, av_stream, None)

        info = StreamVideoInfo(
            context.width,
            context.height,
            context.sample_aspect_ratio.num,
            context.sample_aspect_ratio.den,
            frame_rate.num,
            frame_rate.den,
            context.codec_id,
        )
    elif context.codec_type == AVMEDIA_TYPE_AUDIO:
        # Format changed in 7.x
        if avformat_version >= 61:
            channel_count = context.ch_layout.nb_channels
        else:
            channel_count = context.channels

        if _debug:
            print(" codec_type=", context.codec_type)
            print(" codec_id=", context.codec_id)
            print(" codec name=", avcodec.avcodec_get_name(context.codec_id).decode('utf-8'))
            print(" format=", context.format)
            print(" bit_rate=", context.bit_rate)
            print(" sample_rate=", context.sample_rate)
            print(" channel_count=", channel_count)

        info = StreamAudioInfo(
            context.format,
            context.sample_rate,
            channel_count,
        )
    else:
        return None
    return info


def ffmpeg_open_stream(file: FFmpegFile, index: int) -> FFmpegStream:
    if not 0 <= index < file.context.contents.nb_streams:
        raise FFmpegException(f'index out of range. Only {file.context.contents.nb_streams} streams.')
    codec_context = avcodec.avcodec_alloc_context3(None)
    if not codec_context:
        raise MemoryError('Could not allocate Codec Context.')
    result = avcodec.avcodec_parameters_to_context(
        codec_context,
        file.context.contents.streams[index].contents.codecpar)
    if result < 0:
        avcodec.avcodec_free_context(byref(codec_context))
        raise FFmpegException('Could not copy the AVCodecContext.')
    codec_id = codec_context.contents.codec_id
    codec = avcodec.avcodec_find_decoder(codec_id)
    if _debug:
        print("Found Codec=", codec_id, "=", codec.contents.long_name.decode())

    # VP8 and VP9 default codec don't support alpha transparency.
    # Force libvpx codec in this case.
    if codec_id == AV_CODEC_ID_VP9:
        newcodec = avcodec.avcodec_find_decoder_by_name(b"libvpx-vp9")
        codec = newcodec or codec

    if codec_id == AV_CODEC_ID_VP8:
        newcodec = avcodec.avcodec_find_decoder_by_name(b"libvpx")
        codec = newcodec or codec

    if not codec:
        raise FFmpegException('No codec found for this media. '
                              f'codecID={codec_id}')

    codec_id = codec.contents.id
    if _debug:
        print("Loaded codec: ", codec.contents.long_name.decode())

    result = avcodec.avcodec_open2(codec_context, codec, None)
    if result < 0:
        raise FFmpegException('Could not open the media with the codec.')

    stream = FFmpegStream()
    stream.format_context = file.context
    stream.codec_context = codec_context
    stream.type = codec_context.contents.codec_type
    stream.frame = avutil.av_frame_alloc()
    stream.time_base = file.context.contents.streams[index].contents.time_base

    return stream


def ffmpeg_close_stream(stream: FFmpegStream) -> None:
    if stream.frame:
        avutil.av_frame_free(byref(stream.frame))
    avcodec.avcodec_free_context(byref(stream.codec_context))


def ffmpeg_seek_file(file: FFmpegFile, timestamp: float) -> None:
    flags = 0
    max_ts = file.context.contents.duration * AV_TIME_BASE
    result = avformat.avformat_seek_file(
        file.context, -1, 0,
        timestamp, timestamp, flags,
    )
    if result < 0:
        buf = create_string_buffer(128)
        avutil.av_strerror(result, buf, 128)
        descr = buf.value
        raise FFmpegException('Error occurred while seeking. ' +
                              descr.decode())


def ffmpeg_read(file: FFmpegFile, packet: _Pointer[AVPacket]) -> bool:
    """Read from the stream a packet.

    Returns:
        True if the packet was correctly read. False if the end of stream was reached, or an error occurred.
    """
    avcodec.av_packet_unref(packet)
    result = avformat.av_read_frame(file.context, packet)
    if result < 0:
        return False

    return True


def ffmpeg_get_packet_pts(file: FFmpegFile, packet: _Pointer[AVPacket]) -> float:
    if packet.contents.dts != AV_NOPTS_VALUE:
        pts = packet.contents.dts
    else:
        pts = 0

    timestamp = avutil.av_rescale_q(pts,
                                    file.context.contents.streams[
                                        packet.contents.stream_index].contents.time_base,
                                    AV_TIME_BASE_Q)
    return timestamp


def ffmpeg_get_frame_ts(stream: FFmpegStream) -> float:
    ts = stream.frame.contents.best_effort_timestamp
    timestamp = avutil.av_rescale_q(ts,
                                    stream.time_base,
                                    AV_TIME_BASE_Q)
    return timestamp


def ffmpeg_init_packet() -> _Pointer[AVPacket]:
    p = avcodec.av_packet_alloc()
    if not p:
        raise MemoryError("Could not allocate AVPacket.")
    return p


def ffmpeg_free_packet(packet):
    avcodec.av_packet_free(byref(packet))


def ffmpeg_unref_packet(packet):
    avcodec.av_packet_unref(byref(packet))


def ffmpeg_transfer_packet(dst, src):
    avcodec.av_packet_move_ref(dst, src)


def get_version():
    """Return an informative version string of FFmpeg"""
    return avutil.av_version_info().decode()


def timestamp_from_ffmpeg(timestamp):
    return float(timestamp) / 1000000


def timestamp_to_ffmpeg(timestamp):
    return int(timestamp * 1000000)


class _Packet:
    packet: AVPacket
    timestamp: float

    __slots__ = ['packet', 'timestamp']

    def __init__(self, packet, timestamp):
        self.packet = AVPacket()
        ffmpeg_transfer_packet(byref(self.packet), packet)
        self.timestamp = timestamp

    def __del__(self):
        if ffmpeg_unref_packet is not None:
            ffmpeg_unref_packet(self.packet)


class VideoPacket(_Packet):
    id: int
    image: int
    _next_id = 0

    def __init__(self, packet, timestamp):
        super().__init__(packet, timestamp)
        # Decoded image.  0 == not decoded yet; None == Error or discarded
        self.image = 0
        self.id = self._next_id
        VideoPacket._next_id += 1


class AudioPacket(_Packet):
    pass


class FFmpegSource(StreamingSource):
    # Max increase/decrease of original sample size
    SAMPLE_CORRECTION_PERCENT_MAX = 10

    # Maximum amount of packets to create for video and audio queues.
    MAX_QUEUE_SIZE = 100

    def __init__(self, filename: str, file: BinaryIO | None=None):
        self._packet = None
        self._video_stream = None
        self._audio_stream = None
        self._stream_end = False
        self._file = None
        self._memory_file = None

        encoded_filename = filename.encode(sys.getfilesystemencoding())

        if file:
            self._file, self._memory_file = ffmpeg_open_memory_file(encoded_filename, file)
        else:
            self._file = ffmpeg_open_filename(encoded_filename)

        if not self._file:
            raise FFmpegException(f'Could not open "{filename}"')

        self._video_stream_index = None
        self._audio_stream_index = None
        self._audio_format = None

        self.img_convert_ctx = POINTER(SwsContext)()
        self.audio_convert_ctx = None

        file_info = ffmpeg_file_info(self._file)

        self.info = SourceInfo()
        self.info.title = file_info.title
        self.info.author = file_info.author
        self.info.copyright = file_info.copyright
        self.info.comment = file_info.comment
        self.info.album = file_info.album
        self.info.year = file_info.year
        self.info.track = file_info.track
        self.info.genre = file_info.genre

        # Pick the first video and audio streams found, ignore others.
        for i in range(file_info.n_streams):
            info = ffmpeg_stream_info(self._file, i)

            if isinstance(info, StreamVideoInfo) and self._video_stream is None:
                stream = ffmpeg_open_stream(self._file, i)

                self.video_format = VideoFormat(
                    width=info.width,
                    height=info.height)
                if info.sample_aspect_num != 0:
                    self.video_format.sample_aspect = (
                            float(info.sample_aspect_num) /
                            info.sample_aspect_den)
                self.video_format.frame_rate = (
                        float(info.frame_rate_num) /
                        info.frame_rate_den)
                self._video_stream = stream
                self._video_stream_index = i

            elif isinstance(info, StreamAudioInfo) and self._audio_stream is None:
                stream = ffmpeg_open_stream(self._file, i)

                channels_out = min(2, info.channels)
                channel_input = self._get_default_channel_layout(info.channels)
                channel_output = self._get_default_channel_layout(channels_out)

                sample_bits = info.sample_bits
                if info.sample_format in (AV_SAMPLE_FMT_U8, AV_SAMPLE_FMT_U8P):
                    self.tgt_format = AV_SAMPLE_FMT_U8
                else:
                    # No matter the input format, produce S16 samples.
                    sample_bits = 16
                    self.tgt_format = AV_SAMPLE_FMT_S16

                self.audio_format = AudioFormat(
                    channels=channels_out,
                    sample_size=sample_bits,
                    sample_rate=info.sample_rate)
                self._audio_stream = stream
                self._audio_stream_index = i

                self.audio_convert_ctx = self.get_formatted_swr_context(channel_output, info.sample_rate, channel_input, info.sample_format)
                if not self.audio_convert_ctx:
                    swresample.swr_free(self.audio_convert_ctx)
                    raise FFmpegException('Cannot create sample rate converter.')

                result = swresample.swr_init(self.audio_convert_ctx)
                if result < 0:
                    swresample.swr_free(self.audio_convert_ctx)
                    raise FFmpegException('Cannot create sample rate converter.', result)

        self._packet = ffmpeg_init_packet()
        self._events = []  # They don't seem to be used!

        self.audioq = deque()
        # Make queue big enough to accommodate 1.2 sec?
        self._max_len_audioq = self.MAX_QUEUE_SIZE  # Need to figure out a correct amount
        if self.audio_format:
             # Buffer 1 sec worth of audio
             nbytes = ffmpeg_get_audio_buffer_size(self.audio_format)
             self._audio_buffer = (c_uint8 * nbytes)()

        self.videoq: deque[VideoPacket] = deque()
        self._max_len_videoq = self.MAX_QUEUE_SIZE  # Need to figure out a correct amount

        self.start_time = self._get_start_time()
        self._duration = timestamp_from_ffmpeg(file_info.duration)
        self._duration -= self.start_time

        # Flag to determine if the _fillq method was already scheduled
        self._fillq_scheduled = False
        self._fillq()
        # Don't understand why, but some files show that seeking without
        # reading the first few packets results in a seeking where we lose
        # many packets at the beginning.
        # We only seek back to 0 for media which have a start_time > 0
        if self.start_time > 0:
            self.seek(0.0)

    def get_formatted_swr_context(self, channel_output: AVChannelLayout | int, sample_rate: int,
                                  channel_input: AVChannelLayout | int, sample_format: int) -> int | SwrContext:
        # Newer FFmpeg versions use the AVChannelLayout
        if swresample_version < 5:
            return swresample.swr_alloc_set_opts(None,
                                          channel_output, self.tgt_format, sample_rate,
                                          channel_input, sample_format, sample_rate,
                                          0, None)
        else:
            swr_ctx = swresample.swr_alloc()
            if not swr_ctx:
                raise RuntimeError("Could not allocate SwrContext")

            if swresample.swr_alloc_set_opts2(byref(swr_ctx),
                                              channel_output, self.tgt_format, sample_rate,
                                              channel_input, sample_format, sample_rate,
                                              0, None) < 0:
                raise Exception("Could not set sample rate context values.")
            return swr_ctx

    @staticmethod
    def _get_default_channel_layout(channels: int) -> int | AVChannelLayout:
        if swresample_version < 5:
            return avutil.av_get_default_channel_layout(channels)
        else:
            layout = AVChannelLayout()
            avutil.av_channel_layout_default(byref(layout), channels)
            return layout

    def __del__(self) -> None:
        if self._packet and ffmpeg_free_packet is not None:
            ffmpeg_free_packet(self._packet)
        if self._video_stream and swscale is not None:
            swscale.sws_freeContext(self.img_convert_ctx)
            ffmpeg_close_stream(self._video_stream)
        if self._audio_stream:
            if swresample:
                swresample.swr_free(self.audio_convert_ctx)
                ffmpeg_close_stream(self._audio_stream)
        if self._file and ffmpeg_close_file is not None:
            ffmpeg_close_file(self._file)

    def seek(self, timestamp: float) -> None:
        if _debug:
            print('FFmpeg seek', timestamp)

        ffmpeg_seek_file(
            self._file,
            timestamp_to_ffmpeg(timestamp + self.start_time),
        )
        del self._events[:]
        self._stream_end = False
        self._clear_video_audio_queues()
        self._fillq()

        # Consume video and audio packets until we arrive at the correct
        # timestamp location
        if not self.audio_format:
            while len(self.videoq) > 1:
                # We only advance if there is at least 2 packets in the queue
                # The queue is only left with 1 packet if we have reached the
                #  end of the stream.
                if timestamp < self.videoq[1].timestamp:
                    break
                else:
                    self.get_next_video_frame(skip_empty_frame=False)

        elif not self.video_format:
            while len(self.audioq) > 1:
                # We only advance if there is at least 2 packets in the queue
                # The queue is only left with 1 packet if we have reached the
                #  end of the stream.
                if timestamp < self.audioq[1].timestamp:
                    break
                else:
                    self._get_audio_packet()

        else:
            while len(self.audioq) > 1 and len(self.videoq) > 1:
                # We only advance if there is at least 2 packets in the queue
                # The queue is only left with 1 packet if we have reached the
                #  end of the stream.
                audioq_is_first = self.audioq[0].timestamp < self.videoq[0].timestamp
                correct_audio_pos = timestamp < self.audioq[1].timestamp
                correct_video_pos = timestamp < self.videoq[1].timestamp
                if audioq_is_first and not correct_audio_pos:
                    self._get_audio_packet()

                elif not correct_video_pos:
                    self.get_next_video_frame(skip_empty_frame=False)

                else:
                    break

    def _get_audio_packet(self) -> AudioPacket:
        """Take an audio packet from the queue.

        This function will schedule its `_fillq` function to fill up
        the queues if space is available. Multiple calls to this method will
        only result in one scheduled call to `_fillq`.
        """
        audio_data = self.audioq.popleft()
        low_lvl = self._check_low_level()
        if not low_lvl and not self._fillq_scheduled:
            pyglet.clock.schedule_once(lambda dt: self._fillq(), 0)
            self._fillq_scheduled = True

        return audio_data

    def _get_video_packet(self) -> VideoPacket | None:
        """Take a video packet from the queue.

        This function will schedule its `_fillq` function to fill up
        the queues if space is available. Multiple calls to this method will
        only result in one scheduled call to `_fillq`.
        """
        if not self.videoq:
            return None
        video_packet = self.videoq.popleft()
        low_lvl = self._check_low_level()
        if not low_lvl and not self._fillq_scheduled:
            pyglet.clock.schedule_once(lambda dt: self._fillq(), 0)
            self._fillq_scheduled = True
        return video_packet

    def _clear_video_audio_queues(self) -> None:
        """Empty both audio and video queues."""
        self.audioq.clear()
        self.videoq.clear()

    def _fillq(self) -> None:
        """Fill up both Audio and Video queues if space is available in both"""
        # We clear our flag.
        self._fillq_scheduled = False
        while (len(self.audioq) < self._max_len_audioq and
               len(self.videoq) < self._max_len_videoq):
            if self._get_packet():
                self._process_packet()
            else:
                self._stream_end = True
                break

    def _check_low_level(self) -> bool:
        """Check if both audio and video queues are getting very low.

        If one of them has less than 2 elements, we fill the queue immediately
        with new packets. We don't wait for a scheduled call because we need
        them immediately.

        This would normally happens only during seek operations where we
        consume many packets to find the correct timestamp.
        """
        if len(self.audioq) < 2 or len(self.videoq) < 2:
            if len(self.audioq) < self._max_len_audioq and len(self.videoq) < self._max_len_videoq:
                self._fillq()
            return True
        return False

    def _get_packet(self) -> bool:
        # Read a packet into self._packet. Returns True if OK, False if no
        # more packets are in stream.
        return ffmpeg_read(self._file, self._packet)

    def _process_packet(self) -> AudioPacket | VideoPacket:
        """Process the packet that has been just read.

        Determines whether it's a video or audio packet and queue it in the
        appropriate queue.
        """
        timestamp = ffmpeg_get_packet_pts(self._file, self._packet)
        timestamp = timestamp_from_ffmpeg(timestamp)
        timestamp -= self.start_time

        if self._packet.contents.stream_index == self._video_stream_index:
            video_packet = VideoPacket(self._packet, timestamp)

            if _debug:
                print('Created and queued packet %d (%f)' % (video_packet.id, video_packet.timestamp))

            self.videoq.append(video_packet)
            return video_packet

        elif self.audio_format and self._packet.contents.stream_index == self._audio_stream_index:
            audio_packet = AudioPacket(self._packet, timestamp)

            self.audioq.append(audio_packet)
            return audio_packet

    def get_audio_data(self, num_bytes: int, compensation_time: float=0.0) -> AudioData | None:
        data = b''
        timestamp = duration = 0

        while len(data) < num_bytes:
            if not self.audioq:
                break

            audio_packet = self._get_audio_packet()
            buffer, timestamp, duration = self._decode_audio_packet(audio_packet, compensation_time)

            if not buffer:
                break
            data += buffer

        # No data and no audio queue left
        if not data and not self.audioq:
            if not self._stream_end:
                # No more audio data in queue, but we haven't hit the stream end.
                if _debug:
                    print("Audio queue was starved by the audio driver.")

            return None

        audio_data = AudioData(data, len(data), timestamp, duration, [])

        while self._events and self._events[0].timestamp <= (timestamp + duration):
            event = self._events.pop(0)
            if event.timestamp >= timestamp:
                event.timestamp -= timestamp
                audio_data.events.append(event)

        if _debug:
            print(f'get_audio_data returning ts {audio_data.timestamp} with events {audio_data.events}')
            print('remaining events are', self._events)

        return audio_data

    def _decode_audio_packet(self, audio_packet: AudioPacket, compensation_time: float) -> tuple[bytes | None, float, float]:
        while True:
            try:
                size_out = self._ffmpeg_decode_audio(
                    audio_packet.packet,
                    self._audio_buffer,
                    compensation_time)
            except FFmpegException:
                break

            if size_out <= 0:
                break

            buffer = create_string_buffer(size_out)
            memmove(buffer, self._audio_buffer, len(buffer))
            buffer = buffer.raw

            duration = float(len(buffer)) / self.audio_format.bytes_per_second
            timestamp = ffmpeg_get_frame_ts(self._audio_stream)
            timestamp = timestamp_from_ffmpeg(timestamp)
            return buffer, timestamp, duration

        return None, 0, 0

    def _ffmpeg_decode_audio(self, packet: AVPacket, data_out: Array[c_uint8], compensation_time: float) -> int:
        stream = self._audio_stream

        if stream.type != AVMEDIA_TYPE_AUDIO:
            raise FFmpegException('Trying to decode audio on a non-audio stream.')

        sent_result = avcodec.avcodec_send_packet(
            stream.codec_context,
            packet,
        )

        if sent_result < 0:
            buf = create_string_buffer(128)
            avutil.av_strerror(sent_result, buf, 128)
            descr = buf.value
            raise FFmpegException(f'Error occurred sending packet to decoder. {descr.decode()}')

        receive_result = avcodec.avcodec_receive_frame(
            stream.codec_context,
            stream.frame,
        )

        if receive_result < 0:
            buf = create_string_buffer(128)
            avutil.av_strerror(receive_result, buf, 128)
            descr = buf.value
            raise FFmpegException(f'Error occurred receiving frame. {descr.decode()}')

        plane_size = c_int(0)

        if avformat_version >= 61:
            channel_count = stream.codec_context.contents.ch_layout.nb_channels
        else:
            channel_count = stream.codec_context.contents.channels

        data_size = avutil.av_samples_get_buffer_size(
            byref(plane_size),
            channel_count,
            stream.frame.contents.nb_samples,
            stream.codec_context.contents.sample_fmt,
            1)

        if data_size < 0:
            raise FFmpegException('Error in av_samples_get_buffer_size')
        if len(self._audio_buffer) < data_size:
            raise FFmpegException('Output audio buffer is too small for current audio frame!')

        nb_samples = stream.frame.contents.nb_samples
        sample_rate = stream.codec_context.contents.sample_rate
        bytes_per_sample = avutil.av_get_bytes_per_sample(self.tgt_format)
        channels_out = min(2, self.audio_format.channels)

        wanted_nb_samples = nb_samples + compensation_time * sample_rate
        min_nb_samples = (nb_samples * (100 - self.SAMPLE_CORRECTION_PERCENT_MAX) / 100)
        max_nb_samples = (nb_samples * (100 + self.SAMPLE_CORRECTION_PERCENT_MAX) / 100)
        wanted_nb_samples = min(max(wanted_nb_samples, min_nb_samples), max_nb_samples)
        wanted_nb_samples = int(wanted_nb_samples)

        if wanted_nb_samples != nb_samples:
            res = swresample.swr_set_compensation(
                self.audio_convert_ctx,
                (wanted_nb_samples - nb_samples),
                wanted_nb_samples,
            )

            if res < 0:
                raise FFmpegException('swr_set_compensation failed.')

        data_in = stream.frame.contents.extended_data
        p_data_out = cast(data_out, POINTER(c_uint8))

        out_samples = swresample.swr_get_out_samples(self.audio_convert_ctx, nb_samples)
        total_samples_out = swresample.swr_convert(self.audio_convert_ctx,
                                                   byref(p_data_out), out_samples,
                                                   data_in, nb_samples)
        while True:
            # We loop because there could be some more samples buffered in
            # SwrContext. We advance the pointer where we write our samples.
            offset = (total_samples_out * channels_out * bytes_per_sample)
            p_data_offset = cast(
                addressof(p_data_out.contents) + offset,
                POINTER(c_uint8),
            )

            samples_out = swresample.swr_convert(self.audio_convert_ctx,
                                                 byref(p_data_offset),
                                                 out_samples - total_samples_out, None, 0)
            if samples_out == 0:
                # No more samples. We can continue.
                break
            total_samples_out += samples_out

        size_out = (total_samples_out * channels_out * bytes_per_sample)

        return size_out

    def _decode_video_packet(self, video_packet: VideoPacket) -> None:
        # # Some timing and profiling
        # pr = cProfile.Profile()
        # pr.enable()
        # clock = pyglet.clock.get_default()
        # t0 = clock.time()

        width = self.video_format.width
        height = self.video_format.height
        pitch = width * 4
        # https://ffmpeg.org/doxygen/3.3/group__lavc__decoding.html#ga8f5b632a03ce83ac8e025894b1fc307a
        nbytes = (pitch * height + FF_INPUT_BUFFER_PADDING_SIZE)
        buffer = (c_uint8 * nbytes)()
        try:
            result = self._ffmpeg_decode_video(video_packet.packet, buffer)
        except FFmpegException:
            image_data = None
        else:
            image_data = image.ImageData(width, height, 'RGBA', buffer, pitch)
            timestamp = ffmpeg_get_frame_ts(self._video_stream)
            timestamp = timestamp_from_ffmpeg(timestamp)
            video_packet.timestamp = timestamp - self.start_time

        video_packet.image = image_data

        if _debug:
            print('Decoding video packet at timestamp', video_packet, video_packet.timestamp)

            # t2 = clock.time()
            # pr.disable()
            # print("Time in _decode_video_packet: {:.4f} s for timestamp {} s".format(t2-t0, packet.timestamp))
            # if t2-t0 > 0.01:
            #     import pstats
            #     ps = pstats.Stats(pr).sort_stats("cumulative")
            #     ps.print_stats()

    def _ffmpeg_decode_video(self, packet: AVPacket, data_out: Array[c_uint8]) -> int:
        stream = self._video_stream
        rgba_ptrs = (POINTER(c_uint8) * 4)()
        rgba_stride = (c_int * 4)()
        width = stream.codec_context.contents.width
        height = stream.codec_context.contents.height
        if stream.type != AVMEDIA_TYPE_VIDEO:
            raise FFmpegException('Trying to decode video on a non-video stream.')

        sent_result = avcodec.avcodec_send_packet(
            stream.codec_context,
            packet,
        )

        if sent_result < 0:
            buf = create_string_buffer(128)
            avutil.av_strerror(sent_result, buf, 128)
            descr = buf.value
            raise FFmpegException(f'Video: Error occurred sending packet to decoder. {descr.decode()}')

        receive_result = avcodec.avcodec_receive_frame(
            stream.codec_context,
            stream.frame,
        )

        if receive_result < 0:
            buf = create_string_buffer(128)
            avutil.av_strerror(receive_result, buf, 128)
            descr = buf.value
            raise FFmpegException(f'Video: Error occurred receiving frame. {descr.decode()}')

        avutil.av_image_fill_arrays(rgba_ptrs, rgba_stride, data_out,
                                    AV_PIX_FMT_RGBA, width, height, 1)

        self.img_convert_ctx = swscale.sws_getCachedContext(
            self.img_convert_ctx,
            width, height, stream.codec_context.contents.pix_fmt,
            width, height, AV_PIX_FMT_RGBA,
            SWS_FAST_BILINEAR, None, None, None)

        swscale.sws_scale(self.img_convert_ctx,
                          cast(stream.frame.contents.data,
                               POINTER(POINTER(c_uint8))),
                          stream.frame.contents.linesize,
                          0,
                          height,
                          rgba_ptrs,
                          rgba_stride)
        return receive_result

    def get_next_video_timestamp(self) -> int | None:
        if not self.video_format:
            return None

        ts = None

        if self.videoq:
            while True:
                # We skip video packets which are not video frames
                # This happens in mkv files for the first few frames.
                try:
                    video_packet = self.videoq.popleft()
                except IndexError:
                    break
                if video_packet.image == 0:
                    self._decode_video_packet(video_packet)
                if video_packet.image is not None:
                    ts = video_packet.timestamp
                    self.videoq.appendleft(video_packet)  # put it back
                    break
                self._get_video_packet()
        else:
            ts = None

        if _debug:
            print('Next video timestamp is', ts)
        return ts

    def get_next_video_frame(self, skip_empty_frame: bool=True) -> int | None:
        if not self.video_format:
            return None

        while True:
            # We skip video packets which are not video frames
            # This happens in mkv files for the first few frames.
            video_packet = self._get_video_packet()
            if not video_packet:
                return None
            if video_packet.image == 0:
                self._decode_video_packet(video_packet)
            if video_packet.image is not None or not skip_empty_frame:
                break

        if _debug:
            print('Returning', video_packet)

        return video_packet.image

    def _get_start_time(self) -> int:
        def streams() -> Iterator[AVStream]:
            format_context = self._file.context
            for idx in (self._video_stream_index, self._audio_stream_index):
                if idx is None:
                    continue
                stream = format_context.contents.streams[idx].contents
                yield stream

        def start_times(streams) -> Iterator[int]:
            yield 0
            for stream in streams:
                start = stream.start_time
                if start == AV_NOPTS_VALUE:
                    yield 0
                start_time = avutil.av_rescale_q(start,
                                                 stream.time_base,
                                                 AV_TIME_BASE_Q)
                start_time = timestamp_from_ffmpeg(start_time)
                yield start_time

        return max(start_times(streams()))

    @property
    def audio_format(self) -> AudioFormat:
        return self._audio_format

    @audio_format.setter
    def audio_format(self, value: AudioFormat) -> None:
        self._audio_format = value
        if value is None:
            self.audioq.clear()


ffmpeg_init()
if pyglet.options['debug_media']:
    _debug = True
else:
    _debug = False
    avutil.av_log_set_level(8)


#########################################
#   Decoder class:
#########################################

class FFmpegDecoder(MediaDecoder):

    def get_file_extensions(self) -> Sequence[str]:
        return '.mp3', '.ogg'

    def decode(self, filename: str, file: BinaryIO | None, streaming: bool=True) -> FFmpegSource | StaticSource:
        if streaming:
            return FFmpegSource(filename, file)
        else:
            return StaticSource(FFmpegSource(filename, file))


def get_decoders() -> list[FFmpegDecoder]:
    return [FFmpegDecoder()]


def get_encoders() -> list:
    return []



"""
Responsibilities

Path building for entities into a session should be delegated to fs.PathServices
Session's creation, activation and management at start of fs
Versions capture are handled at start of module fs
Utility functions to load - save at the end of fs

See directory layout in manual.
"""
import sys
import json
import os
import pickle
import shutil
import subprocess

import mpexceptions

def get_media_player_path():
    here = os.path.basename(os.path.abspath(__file__))
    path = os.path.join("..", "..", "examples")
    mp_path = os.path.abspath(path)
    return mp_path

# available sessions, activation related functionality #######################


def activation_set_to(session):
    """
    exceptions:
        mpexceptions.ExceptionUndefinedSamplesDir
        mpexceptions.ExceptionNoSessionWithThatName
        mpexceptions.ExceptionSamplesDirDoesNotExist
    """
    samples_dir = get_current_samples_dir()
    activation_file = get_activation_filename(samples_dir)
    if session is None:
        if os.path.isfile(activation_file):
            os.remove(activation_file)
    else:
        json_save(session, activation_file)


def get_current_samples_dir():
    if "pyglet_mp_samples_dir" not in os.environ:
        raise mpexceptions.ExceptionUndefinedSamplesDir()
    path = os.environ["pyglet_mp_samples_dir"]
    if not os.path.isdir(path):
        raise mpexceptions.ExceptionSamplesDirDoesNotExist(path)
    return path


def get_path_info_for_active_session():
    samples_dir = get_current_samples_dir()
    activation_file = get_activation_filename(samples_dir)
    try:
        session = json_load(activation_file)
    except FileNotFoundError:
        raise mpexceptions.ExceptionNoSessionIsActive()
    pathserv = PathServices(samples_dir, session)
    return pathserv


def get_path_info_for_session(session):
    """
    samples_dir comes from env var pyglet_mp_samples_dir
    """
    samples_dir = get_current_samples_dir()
    pathserv = PathServices(samples_dir, session)
    conf_file = pathserv.configuration_filename()
    if os.path.isdir(pathserv.session_dir) and os.path.isfile(conf_file):
        return pathserv
    raise mpexceptions.ExceptionNoSessionWithThatName()


def get_sessions(samples_dir):
    candidates = [s for s in os.listdir(samples_dir)
                  if os.path.isdir(os.path.join(samples_dir, s))]
    sessions = []
    for name in candidates:
        pathserv = PathServices(samples_dir, name)
        if os.path.isfile(pathserv.configuration_filename()):
            sessions.append(name)
    return sessions


def get_activation_filename(samples_dir):
    return os.path.join(samples_dir, "activation.json")


def new_session(session, playlist_file=None):
    """creates a new session and sets as the active session"""
    samples_dir = get_current_samples_dir()
    new_session_for_samples_dir(session, samples_dir, playlist_file)
    activation_set_to(session)


def new_session_for_samples_dir(session, samples_dir, playlist_file=None):
    """creates a new session and returns it's pathserv, session is not activated

    Does not use env vars.
    """
    if not os.path.isdir(samples_dir):
        raise mpexceptions.ExceptionSamplesDirDoesNotExist(samples_dir)
    pathserv = PathServices(samples_dir, session)
    if os.path.exists(pathserv.session_dir):
        raise mpexceptions.ExceptionSessionExistWithSameName()
    conf = default_initial_configuration()

    if playlist_file is not None:
        if not os.path.isfile(playlist_file):
            raise mpexceptions.ExceptionPlaylistFileDoNotExists()
        conf["has_playlist"] = True
        samples = sane_samples_from_playlist(pathserv, playlist_file)
    else:
        samples = [sample for sample, fname in pathserv.playlist_generator_all()]

    # directory creation delayed to last possible moment so exceptions don't
    # left empty session dir
    for path in pathserv.dirs_to_create():
        os.mkdir(path)

    dump_pyglet_info(pathserv)
    copy_samples_version_to_dbg(pathserv)
    save_session_playlist(pathserv, samples)
    json_save(conf, pathserv.configuration_filename())
    dump_hg_changeset(pathserv)
    return pathserv


def dump_pyglet_info(pathserv):
    import pyglet
    import pyglet.info
    filename = pathserv.special_raw_filename("pyglet_info")
    old = sys.stdout
    try:
        with open(filename, "w", encoding="utf-8") as f:
            sys.stdout = f
            pyglet.info.dump()
    except Exception as ex:
        import traceback
        traceback.print_exc()
    finally:
        sys.stdout = old


# assumes the cwd hits into pyglet clone under test
def dump_hg_changeset(pathserv):
    fname = pathserv.special_raw_filename("pyglet_hg_revision")
    # win needs shell=True to locate the 'hg'
    shell = (sys.platform == "win32")
    with open(fname, "w", encoding="utf8") as outfile:
        subprocess.call(["hg", "parents", "--encoding", "utf8"],
                    shell=shell,
                    stdout=outfile,
                    timeout=5.0)


def copy_samples_version_to_dbg(pathserv):
    src = pathserv.samples_version_filename()
    dst = pathserv.special_raw_filename("samples_version")
    shutil.copyfile(src, dst)


def sane_samples_from_playlist(pathserv, playlist_file):
    # files exist and extension is not blacklisted
    samples = []
    rejected = []
    for sample, filename in pathserv.playlist_generator_from_file(playlist_file):
        ext = os.path.splitext(sample)[1]
        if (ext in {".dbg", ".htm", ".html", ".json",
                    ".log", ".pkl", ".py", ".txt"} or
            not os.path.isfile(filename)):
            rejected.append(sample)
        else:
            samples.append(sample)
    if len(rejected) > 0:
        raise mpexceptions.ExceptionBadSampleInPlaylist(rejected)
    return samples


def save_session_playlist(pathserv, samples):
    outfile = pathserv.special_raw_filename("session_playlist")
    text = "\n".join(sorted(samples))
    txt_save(text, outfile)


def load_session_playlist(pathserv):
    infile = pathserv.special_raw_filename("session_playlist")
    text = txt_load(infile)
    lines = text.split("\n")
    samples = [s.strip() for s in lines]
    return samples


# session configuration functionality #########################################


def default_initial_configuration():
    conf = {
        "config_vs": "1.0",
        "protect_raw_data": False,
        "protect_reports": False,
        "dev_debug": False,
        "has_playlist": False,
        }
    return conf


def update_active_configuration(modify):
    """updates the active session configuration file with the key: values in modify"""
    pathserv = get_path_info_for_active_session()
    update_configuration(pathserv, modify)


# validation over modify keys and values is a caller responsibility,
# here only verified that if key in modify then key in conf
def update_configuration(pathserv, modify):
    """The session associated with pathserv gets its configuration updated
    with the key: values in modify"""
    conf = json_load(pathserv.configuration_filename())
    for k in modify:
        assert k in conf
    conf.update(modify)
    json_save(conf, pathserv.configuration_filename())


def get_session_configuration(pathserv):
    conf = json_load(pathserv.configuration_filename())
    return conf


# path functionality according to path schema ##################################


class PathServices:
    def __init__(self, samples_dir, session):
        self.samples_dir = samples_dir
        self.session = session
        self.session_dir = os.path.join(samples_dir, session)

        self.dbg_dir = os.path.join(self.session_dir, "dbg")
        self.rpt_dir = os.path.join(self.session_dir, "reports")

    def dirs_to_create(self):
        return [self.session_dir, self.dbg_dir, self.rpt_dir]

    def sample(self, filename):
        """returns sample name from filename

        dev note: no checks performed, maybe it should check
           - filename point into the sample_dir directory
           - filename exists
        """
        return os.path.basename(filename)

    def filename(self, sample):
        """Returns the filename for the sample

        dev note: no checks performed, maybe it should check
           - filename exists
        """
        return os.path.join(self.samples_dir, sample)

    def dbg_filename(self, sample):
        """Returns filename to store media_player's internal state recording for sample"""
        return os.path.join(self.dbg_dir, sample + ".dbg")

    def sample_from_dbg_filename(self, filename):
        return os.path.splitext(filename)[0]

    def samples_version_filename(self):
        return os.path.join(self.samples_dir, "_version.txt")

    def report_filename(self, sample, report_name, txt=True):
        """Returns filename to store a 'per sample' report for sample"""
        if txt:
            s = os.path.join(self.rpt_dir, "%s.%s.txt" % (sample, report_name))
        else:
            s = os.path.join(self.rpt_dir, "%s.%s" % (sample, report_name))
        return s

    def special_report_filename(self, report_name):
        """Returns filename to store a report that don't depend on a single sample"""
        table = {
            # report_name: shortname
            "summary": "00_summary.txt",
            "pyglet_info": "00_pyglet_info.txt",
            "pyglet_hg_revision": "04_pyglet_hg_revision.txt"
            }
        return os.path.join(self.rpt_dir, table[report_name])

    def special_raw_filename(self, shortname):
        table = {
            "session_playlist": "_session_playlist.txt",
            "crashes_light": "_crashes_light.pkl",
            "pyglet_info": "_pyglet_info.txt",
            "samples_version": "_samples_version.txt",
            "pyglet_hg_revision": "_pyglet_hg_revision.txt"
            }
        return os.path.join(self.dbg_dir, table[shortname])

    def configuration_filename(self):
        return os.path.join(self.session_dir, "configuration.json")

    def playlist_generator(self, playlist_text):
        lines = playlist_text.split("\n")
        for line in lines:
            sample = line.strip()
            if sample == "" or sample.startswith("#"):
                continue
            yield (sample, self.filename(sample))

    def playlist_generator_from_file(self, playlist_file):
        with open(playlist_file, "r", encoding="utf-8") as f:
            text = f.read()
        gen = self.playlist_generator(text)
        yield from gen

    def playlist_generator_all(self):
        for name in os.listdir(self.samples_dir):
            filename = os.path.join(self.samples_dir, name)
            ext = os.path.splitext(name)[1]
            if (ext in {".dbg", ".htm", ".html", ".json",
                        ".log", ".pkl", ".py", ".txt"} or
                os.path.isdir(filename)):
                continue
            yield (name, filename)

    def playlist_generator_from_samples_iterable(self, samples):
        for sample in samples:
            yield (sample, self.filename(sample))

    def playlist_generator_from_fixed_text(self):
        text = "0\n1\n2\n3\n4\n5\n6\n7\n8\n9\nz"
        gen = self.playlist_generator(text)
        yield from gen

    def session_playlist_generator(self):
        playlist_file = self.special_raw_filename("session_playlist")
        gen = self.playlist_generator_from_file(playlist_file)
        yield from gen

    def session_exists(self):
        filename = self.configuration_filename()
        return os.path.isdir(self.session_dir) and os.path.isfile(filename)


# fs io #######################################################################


def json_save(obj, filename):
    with open(filename, "w", encoding="utf-8") as f:
        json.dump(obj, f)


def json_load(filename):
    with open(filename, "r", encoding="utf-8") as f:
        obj = json.load(f)
    return obj


def pickle_save(obj, filename):
    with open(filename, "wb") as f:
        pickle.dump(obj, f)


def pickle_load(filename):
    with open(filename, "rb") as f:
        obj = pickle.load(f)
    return obj


def txt_load(filename):
    with open(filename, "r", encoding="utf-8") as f:
        text = f.read()
    return text


def txt_save(text, filename):
    with open(filename, "w", encoding="utf-8") as f:
        f.write(text)


def test_composition():
    text = "0\n1\n2\n3\n4\n5\n6\n7\n8\n9\nz"
    # samples_dir, session
    args = "aaa", "bbb"
    a = PathServices(*args)
    for u, v in zip(a.playlist_generator(text), a.playlist_generator_from_fixed_text()):
        print(u, v)

#test_composition()



from collections import deque
import math
import threading
from typing import Deque, Tuple, TYPE_CHECKING

from pyglet.media.drivers.base import AbstractAudioDriver, AbstractAudioPlayer, MediaEvent
from pyglet.media.player_worker_thread import PlayerWorkerThread
from pyglet.media.drivers.listener import AbstractListener
from pyglet.util import debug_print
from . import interface

if TYPE_CHECKING:
    from pyglet.media.codecs import AudioData, AudioFormat, Source
    from pyglet.media.player import Player


_debug = debug_print('debug_media')


def _convert_coordinates(coordinates: Tuple[float, float, float]) -> Tuple[float, float, float]:
    x, y, z = coordinates
    return x, y, -z


class XAudio2Driver(AbstractAudioDriver):
    def __init__(self) -> None:
        self._xa2_driver = interface.XAudio2Driver()
        self._xa2_listener = self._xa2_driver.create_listener()
        self._listener = XAudio2Listener(self._xa2_listener, self._xa2_driver)

        self.worker = PlayerWorkerThread()
        self.worker.start()

    def get_performance(self) -> interface.lib.XAUDIO2_PERFORMANCE_DATA:
        assert self._xa2_driver is not None
        return self._xa2_driver.get_performance()

    def create_audio_player(self, source: 'Source', player: 'Player') -> 'XAudio2AudioPlayer':
        assert self._xa2_driver is not None
        return XAudio2AudioPlayer(self, source, player)

    def get_listener(self) -> 'XAudio2Listener':
        return self._listener

    def delete(self) -> None:
        if self._xa2_driver is not None:
            self.worker.stop()
            self.worker = None
            self._xa2_driver._delete_driver()
            self._xa2_driver = None
            self._xa2_listener = None


class XAudio2Listener(AbstractListener):
    def __init__(self, xa2_listener, xa2_driver) -> None:
        self._xa2_listener = xa2_listener
        self._xa2_driver = xa2_driver

    def _set_volume(self, volume: float) -> None:
        self._volume = volume
        self._xa2_driver.volume = volume

    def _set_position(self, position: Tuple[float, float, float]) -> None:
        self._position = position
        self._xa2_listener.position = _convert_coordinates(position)

    def _set_forward_orientation(self, orientation: Tuple[float, float, float]) -> None:
        self._forward_orientation = orientation
        self._set_orientation()

    def _set_up_orientation(self, orientation: Tuple[float, float, float]) -> None:
        self._up_orientation = orientation
        self._set_orientation()

    def _set_orientation(self) -> None:
        self._xa2_listener.orientation = (
            _convert_coordinates(self._forward_orientation) +
            _convert_coordinates(self._up_orientation))


class XAudio2AudioPlayer(AbstractAudioPlayer):
    def __init__(self, driver: 'XAudio2Driver', source: 'Source', player: 'Player') -> None:
        super().__init__(source, player)
        # We keep here a strong reference because the AudioDriver is anyway
        # a singleton object which will only be deleted when the application
        # shuts down. The AudioDriver does not keep a ref to the AudioPlayer.
        self.driver = driver

        # Need to cache these because pyglet API allows update separately, but
        # XAudio2 requires both to be set at once.
        self._cone_inner_angle = 360
        self._cone_outer_angle = 360

        # Desired play state. (`True` doesn't necessarily mean the player is playing.
        # It may be silent due to either underrun or because a flush is in progress.)
        self._playing = False

        # Theoretical write and play cursors for an infinite buffer.  play
        # cursor is always <= write cursor (when equal, underrun is
        # happening).
        self._write_cursor = 0
        self._play_cursor = 0

        self._audio_data_in_use: Deque['AudioData'] = deque()
        self._pyglet_source_exhausted = False

        # A lock to be held whenever modifying things relating to the in-use audio data.
        # Ensures that the XAudio2 callbacks will not interfere with the
        # player operations.
        self._audio_data_lock = threading.Lock()

        self._xa2_source_voice = self.driver._xa2_driver.get_source_voice(source.audio_format, self)

    def on_driver_destroy(self) -> None:
        self.stop()
        self._xa2_source_voice = None

    def on_driver_reset(self) -> None:
        self._xa2_source_voice = self.driver._xa2_driver.get_source_voice(self.source.audio_format, self)

        # Queue up any buffers that are still in queue but weren't deleted. This does not
        # pickup where the last sample played, only where the last buffer was submitted.
        # As such, audio will be replayed.
        # TODO: Make best effort by using XAUDIO2_BUFFER.PlayBegin in conjunction
        # with last playback sample
        for audio_data in self._audio_data_in_use:
            xa2_buffer = interface.create_xa2_buffer(audio_data)
            self._xa2_source_voice.submit_buffer(xa2_buffer)

    def delete(self) -> None:
        if self.driver._xa2_driver is None:
            assert _debug("Xaudio2: Player deleted, driver is gone")
            # Driver was deleted, voice is gone; just break up some references and return
            self.driver = None
            self._xa2_source_voice = None
            self._audio_data_in_use.clear()
            return

        assert _debug("XAudio2: Player deleted, returning voice")

        self.stop()
        self.driver._xa2_driver.return_voice(self._xa2_source_voice, self._audio_data_in_use)
        self.driver = None
        self._xa2_source_voice = None

    def play(self) -> None:
        assert _debug(f'XAudio2 play: {self._playing=}')

        if not self._playing:
            self._playing = True
            self._xa2_source_voice.play()
            self.driver.worker.add(self)

        assert _debug('return XAudio2 play')

    def stop(self) -> None:
        assert _debug('XAudio2 stop')

        if self._playing:
            self.driver.worker.remove(self)
            # no callback could possibly be running after this lock is released.
            with self.driver._xa2_driver.lock:
                self._xa2_source_voice.stop()
            self._playing = False

        assert _debug('return XAudio2 stop')

    def clear(self) -> None:
        assert _debug('XAudio2 clear')
        super().clear()
        self._play_cursor = 0
        self._write_cursor = 0
        self._pyglet_source_exhausted = False
        self.driver._xa2_driver.return_voice(self._xa2_source_voice, self._audio_data_in_use)
        self._audio_data_in_use = deque()

        self._xa2_source_voice = self.driver._xa2_driver.get_source_voice(self.source.audio_format, self)
        self._xa2_source_voice.volume = self.player.volume
        self._xa2_source_voice.frequency = self.player.pitch
        if self._xa2_source_voice.is_emitter:
            self._xa2_source_voice.position = _convert_coordinates(self.player.position)
            self._xa2_source_voice.distance_scaler = self.player.min_distance
            self._xa2_source_voice.cone_orientation = _convert_coordinates(self.player.cone_orientation)
            self._xa2_source_voice.cone_outside_volume = self.player.cone_outer_gain
            self._set_cone_angles()
            self.driver._xa2_driver.apply3d(self._xa2_source_voice)

    def on_buffer_end(self, buffer_context_ptr: int) -> None:
        # Called from the XAudio2 thread.
        # A buffer stopped being played by the voice, it should by all means be the first one
        with self._audio_data_lock:
            assert self._audio_data_in_use
            self._audio_data_in_use.popleft()
            # This should cause the AudioData to lose all its references and be gc'd

            if self._audio_data_in_use:
                assert _debug(f"Buffer ended, others remain: {len(self._audio_data_in_use)=}")
                return

            assert self._xa2_source_voice.buffers_queued == 0

            if self._pyglet_source_exhausted:
                # Last buffer ran out naturally, out of AudioData; voice will now fall silent
                assert _debug("Last buffer ended normally, dispatching eos")
                MediaEvent('on_eos').sync_dispatch_to_player(self.player)
            else:
                # Shouldn't have ran out; supplier is running behind
                # All we can do is wait; as long as voices are not stopped via `Stop`, they will
                # immediately continue playing the new buffer once it arrives
                assert _debug("Last buffer ended normally, source is lagging behind")

    def _refill(self, refill_size: int) -> None:
        """Get one piece of AudioData and submit it to the voice.
        This method will release the lock around the call to `get_audio_data`,
        so make sure it's held upon calling.
        """
        assert _debug(f"XAudio2: Retrieving new buffer of {refill_size}B")

        self._audio_data_lock.release()
        audio_data = self._get_and_compensate_audio_data(refill_size, self._play_cursor)
        self._audio_data_lock.acquire()

        if audio_data is None:
            assert _debug(f"XAudio2: Source is out of data")
            self._pyglet_source_exhausted = True
            if not self._audio_data_in_use:
                MediaEvent('on_eos').sync_dispatch_to_player(self.player)
            return

        xa2_buffer = interface.create_xa2_buffer(audio_data)
        self._audio_data_in_use.append(audio_data)
        self._xa2_source_voice.submit_buffer(xa2_buffer)
        assert _debug(f"XAudio2: Submitted buffer of size {audio_data.length}B")

        self.append_events(self._write_cursor, audio_data.events)
        self._write_cursor += audio_data.length

    def _update_play_cursor(self) -> None:
        voice = self._xa2_source_voice
        self._play_cursor = (
            (voice.samples_played - voice.samples_played_at_last_recycle) *
            self.source.audio_format.bytes_per_frame
        )

    def get_play_cursor(self) -> int:
        return self._play_cursor

    def work(self) -> None:
        with self._audio_data_lock:
            self._update_play_cursor()
            self.dispatch_media_events(self._play_cursor)
            self._maybe_refill()

    def _maybe_refill(self) -> bool:
        if self._pyglet_source_exhausted:
            return False

        remaining_bytes = self._write_cursor - self._play_cursor
        if remaining_bytes >= self._buffered_data_comfortable_limit:
            return False

        missing_bytes = self._buffered_data_ideal_size - remaining_bytes
        self._refill(self.source.audio_format.align_ceil(missing_bytes))
        return True

    def prefill_audio(self) -> None:
        with self._audio_data_lock:
            self._maybe_refill()

    def set_volume(self, volume: float) -> None:
        self._xa2_source_voice.volume = volume

    def set_position(self, position: Tuple[float, float, float]) -> None:
        if self._xa2_source_voice.is_emitter:
            self._xa2_source_voice.position = _convert_coordinates(position)

    def set_min_distance(self, min_distance: float) -> None:
        """Not a true min distance, but similar effect. Changes CurveDistanceScaler default is 1."""
        if self._xa2_source_voice.is_emitter:
            self._xa2_source_voice.distance_scaler = min_distance

    def set_max_distance(self, max_distance: float) -> None:
        """No such thing built into xaudio2"""
        return

    def set_pitch(self, pitch: float) -> None:
        self._xa2_source_voice.frequency = pitch

    def set_cone_orientation(self, cone_orientation: Tuple[float, float, float]) -> None:
        if self._xa2_source_voice.is_emitter:
            self._xa2_source_voice.cone_orientation = _convert_coordinates(cone_orientation)

    def set_cone_inner_angle(self, cone_inner_angle: float) -> None:
        if self._xa2_source_voice.is_emitter:
            self._cone_inner_angle = int(cone_inner_angle)
            self._set_cone_angles()

    def set_cone_outer_angle(self, cone_outer_angle: float) -> None:
        if self._xa2_source_voice.is_emitter:
            self._cone_outer_angle = int(cone_outer_angle)
            self._set_cone_angles()

    def _set_cone_angles(self) -> None:
        inner = min(self._cone_inner_angle, self._cone_outer_angle)
        outer = max(self._cone_inner_angle, self._cone_outer_angle)
        self._xa2_source_voice.set_cone_angles(math.radians(inner), math.radians(outer))

    def set_cone_outer_gain(self, cone_outer_gain: float) -> None:
        if self._xa2_source_voice.is_emitter:
            self._xa2_source_voice.cone_outside_volume = cone_outer_gain



# -*- coding: latin-1 -*-
import re
import os.path
import pyglet
from pyglet.gl import (
    glEnable,
    glBlendFunc,
    glHint,
    GL_LINE_SMOOTH,
    GL_BLEND,
    GL_SRC_ALPHA,
    GL_ONE_MINUS_SRC_ALPHA,
    GL_LINE_SMOOTH_HINT,
    GL_LINES,
    GL_DONT_CARE,
)
from pyglet.gl.gl_compat import (
    glPushAttrib,
    glPopAttrib,
    glPushMatrix,
    glPopMatrix,
    glScalef,
    glTranslatef,
    GL_ENABLE_BIT,
)

import xml.dom
import xml.dom.minidom


class SmoothLineGroup(pyglet.graphics.Group):
    @staticmethod
    def set_state():
        glPushAttrib(GL_ENABLE_BIT)
        glEnable(GL_LINE_SMOOTH)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glHint(GL_LINE_SMOOTH_HINT, GL_DONT_CARE)

    @staticmethod
    def unset_state():
        glPopAttrib()


class Curve:
    PATH_RE = re.compile(r'([MLHVCSQTAZ])([^MLHVCSQTAZ]+)', re.IGNORECASE)
    INT = r'([+-]?\d+)'
    FLOAT = r'(?:[\s,]*)([+-]?\d+(?:\.\d+)?)'

    HANDLERS = {}

    def handle(self, rx, types, HANDLERS=HANDLERS):
        def register(function):
            HANDLERS[self] = (rx and re.compile(rx), function, types)
            return function

        return register

    def __init__(self, spec, batch):
        self.batch = batch

        self.start = None
        self.current = None
        self.min_x = self.min_y = self.max_x = self.max_y = None

        for cmd, value in self.PATH_RE.findall(spec):
            # print (cmd, value)
            if not cmd:
                continue
            rx, handler, types = self.HANDLERS[cmd.upper()]
            if rx is None:
                handler(self, cmd)
            else:
                v = []
                for fields in rx.findall(value):
                    v.append([types[i](e) for i, e in enumerate(fields)])
                handler(self, cmd, v)

    def _determine_rect(self, x, y):
        y = -y
        if self.min_x is None:
            self.min_x = self.max_x = x
            self.min_y = self.max_y = y
        else:
            if self.min_x > x:
                self.min_x = x
            elif self.max_x < x:
                self.max_x = x
            if self.min_y > y:
                self.min_y = y
            elif self.max_y < y:
                self.max_y = y

    @handle('M', FLOAT * 2, (float, float))
    def moveto(self, cmd, points):
        """Start a new sub-path at the given (x,y) coordinate. M (uppercase)
        indicates that absolute coordinates will follow; m (lowercase)
        indicates that relative coordinates will follow. If a relative moveto
        (m) appears as the first element of the path, then it is treated as a
        pair of absolute coordinates. If a moveto is followed by multiple pairs
        of coordinates, the subsequent pairs are treated as implicit lineto
        commands.

        Parameters are (x y)+
        """
        points = [list(map(float, point)) for point in points]
        # XXX handle relative
        # XXX confirm that we always reset start here
        self.start = self.current = points[0]
        if len(points) > 2:
            self.lineto({'m': 'l', 'M': 'L'}[cmd], points[1:])

    @handle('L', FLOAT * 2, (float, float))
    def lineto(self, cmd, points):
        """Draw a line from the current point to the given (x,y) coordinate
        which becomes the new current point. L (uppercase) indicates that
        absolute coordinates will follow; l (lowercase) indicates that relative
        coordinates will follow. A number of coordinates pairs may be specified
        to draw a polyline. At the end of the command, the new current point is
        set to the final set of coordinates provided.

        Parameters are (x y)+
        """
        l = []
        self._determine_rect(*self.current)
        for point in points:
            cx, cy = self.current
            x, y = list(map(float, point))
            l.extend([cx, -cy])
            l.extend([x, -y])
            self.current = (x, y)
            self._determine_rect(x, y)
        self.batch.add(len(l) // 2, GL_LINES, SmoothLineGroup(), ('v2f', l))

    @handle('H', FLOAT, (float,))
    def horizontal_lineto(self, cmd, xvals):
        """Draws a horizontal line from the current point (cpx, cpy) to (x,
        cpy). H (uppercase) indicates that absolute coordinates will follow; h
        (lowercase) indicates that relative coordinates will follow. Multiple x
        values can be provided (although usually this doesn't make sense). At
        the end of the command, the new current point becomes (x, cpy) for the
        final value of x.

        Parameters are x+
        """
        cx, cy = self.current
        self._determine_rect(*self.current)
        x = float(xvals[-1])
        self.batch.add(2, GL_LINES, None, ('v2f', (cx, -cy, x, -cy)))
        self.current = (x, cy)
        self._determine_rect(x, cy)

    @handle('V', FLOAT, (float,))
    def vertical_lineto(self, cmd, yvals):
        """Draws a vertical line from the current point (cpx, cpy) to (cpx, y).
        V (uppercase) indicates that absolute coordinates will follow; v
        (lowercase) indicates that relative coordinates will follow. Multiple y
        values can be provided (although usually this doesn't make sense). At
        the end of the command, the new current point becomes (cpx, y) for the
        final value of y.

        Parameters are y+
        """
        cx, cy = self.current
        self._determine_rect(*self.current)
        y = float(yvals[-1])
        self.batch.add(2, GL_LINES, None, ('v2f', [cx, -cy, cx, -y]))
        self.current = (cx, y)
        self._determine_rect(cx, y)

    @handle('Z', None, None)
    def closepath(self, cmd):
        """Close the current subpath by drawing a straight line from the
        current point to current subpath's initial point.
        """
        self.batch.add(2, GL_LINES, SmoothLineGroup(), ('v2f', self.current + tuple(self.start)))

    @handle('C', FLOAT * 6, (float,) * 6)
    def curveto(self, cmd, control_points):
        """Draws a cubic Bézier curve from the current point to (x,y) using
        (x1,y1) as the control point at the beginning of the curve and (x2,y2)
        as the control point at the end of the curve. C (uppercase) indicates
        that absolute coordinates will follow; c (lowercase) indicates that
        relative coordinates will follow. Multiple sets of coordinates may be
        specified to draw a polybézier. At the end of the command, the new
        current point becomes the final (x,y) coordinate pair used in the
        polybézier.

        Control points are (x1 y1 x2 y2 x y)+
        """
        l = []
        last = None
        for entry in control_points:
            x1, y1, x2, y2, x, y = list(map(float, entry))
            t = 0
            cx, cy = self.current
            self.last_control = (x2, y2)
            self.current = (x, y)
            x1 *= 3
            x2 *= 3
            y1 *= 3
            y2 *= 3
            while t <= 1.01:
                a = t
                a2 = a ** 2
                a3 = a ** 3
                b = 1 - t
                b2 = b ** 2
                b3 = b ** 3
                px = cx * b3 + x1 * b2 * a + x2 * b * a2 + x * a3
                py = cy * b3 + y1 * b2 * a + y2 * b * a2 + y * a3
                if last is not None:
                    l.extend(last)
                    l.extend((px, -py))
                last = (px, -py)
                self._determine_rect(px, py)
                t += 0.01
        self.batch.add(len(l) // 2, GL_LINES, SmoothLineGroup(), ('v2f', l))

    @handle('S', FLOAT * 4, (float,) * 4)
    def smooth_curveto(self, cmd, control_points):
        """Draws a cubic Bézier curve from the current point to (x,y). The
        first control point is assumed to be the reflection of the second
        control point on the previous command relative to the current point.
        (If there is no previous command or if the previous command was not an
        C, c, S or s, assume the first control point is coincident with the
        current point.) (x2,y2) is the second control point (i.e., the control
        point at the end of the curve). S (uppercase) indicates that absolute
        coordinates will follow; s (lowercase) indicates that relative
        coordinates will follow. Multiple sets of coordinates may be specified
        to draw a polybézier. At the end of the command, the new current point
        becomes the final (x,y) coordinate pair used in the polybézier.

        Control points are (x2 y2 x y)+
        """
        assert self.last_control is not None, 'S must follow S or C'

        l = []
        last = None
        for entry in control_points:
            x2, y2, x, y = list(map(float, entry))

            # Reflect last control point
            cx, cy = self.current
            lcx, lcy = self.last_control
            dx, dy = cx - lcx, cy - lcy
            x1, y1 = cx + dx, cy + dy

            t = 0
            cx, cy = self.current
            self.last_control = (x2, y2)
            self.current = (x, y)

            x1 *= 3
            x2 *= 3
            y1 *= 3
            y2 *= 3
            while t <= 1.01:
                a = t
                a2 = a ** 2
                a3 = a ** 3
                b = 1 - t
                b2 = b ** 2
                b3 = b ** 3
                px = cx * b3 + x1 * b2 * a + x2 * b * a2 + x * a3
                py = cy * b3 + y1 * b2 * a + y2 * b * a2 + y * a3
                if last is not None:
                    l.extend(last)
                    l.extend((px, -py))
                last = (px, -py)
                self._determine_rect(px, py)
                t += 0.01
        # degenerate vertices
        self.batch.add(len(l) // 2, GL_LINES, SmoothLineGroup(), ('v2f', l))

    @handle('Q', FLOAT * 4, (float,) * 4)
    def quadratic_curveto(self, cmd, control_points):
        """Draws a quadratic Bézier curve from the current point to (x,y)
        using (x1,y1) as the control point. Q (uppercase) indicates that
        absolute coordinates will follow; q (lowercase) indicates that
        relative coordinates will follow. Multiple sets of coordinates may
        be specified to draw a polybézier. At the end of the command, the
        new current point becomes the final (x,y) coordinate pair used in
        the polybézier.

        Control points are (x1 y1 x y)+
        """
        raise NotImplementedError('not implemented')

    @handle('T', FLOAT * 2, (float,) * 2)
    def smooth_quadratic_curveto(self, cmd, control_points):
        """Draws a quadratic Bézier curve from the current point to (x,y).
        The control point is assumed to be the reflection of the control
        point on the previous command relative to the current point. (If
        there is no previous command or if the previous command was not a
        Q, q, T or t, assume the control point is coincident with the
        current point.) T (uppercase) indicates that absolute coordinates
        will follow; t (lowercase) indicates that relative coordinates will
        follow. At the end of the command, the new current point becomes
        the final (x,y) coordinate pair used in the polybézier.

        Control points are (x y)+
        """
        raise NotImplementedError('not implemented')

    @handle('A', FLOAT * 3 + INT * 2 + FLOAT * 2, (float,) * 3 + (int,) * 2 + (float,) * 2)
    def elliptical_arc(self, cmd, parameters):
        """Draws an elliptical arc from the current point to (x, y). The
        size and orientation of the ellipse are defined by two radii (rx,
        ry) and an x-axis-rotation, which indicates how the ellipse as a
        whole is rotated relative to the current coordinate system. The
        center (cx, cy) of the ellipse is calculated automatically to
        satisfy the constraints imposed by the other parameters.
        large-arc-flag and sweep-flag contribute to the automatic
        calculations and help determine how the arc is drawn.

        Parameters are (rx ry x-axis-rotation large-arc-flag sweep-flag x y)+
        """
        raise NotImplementedError('not implemented')


class SVG:
    def __init__(self, filename, rect=None):
        self._rect = rect
        self.scale_x = None
        self.scale_y = None
        self.translate_x = None
        self.translate_y = None
        dom = xml.dom.minidom.parse(filename)
        tag = dom.documentElement
        if tag.tagName != 'svg':
            raise ValueError(f'document is <{tag.tagName}> instead of <svg>')

        # generate all the drawing elements
        self.batch = pyglet.graphics.Batch()
        self.objects = []
        for tag in tag.getElementsByTagName('g'):
            for tag in tag.getElementsByTagName('path'):
                self.objects.append(Curve(tag.getAttribute('d'), self.batch))

        # determine drawing bounds
        self.min_x = min(o.min_x for o in self.objects)
        self.max_x = max(o.max_x for o in self.objects)
        self.min_y = min(o.min_y for o in self.objects)
        self.max_y = max(o.max_y for o in self.objects)

        # determine or apply drawing rect
        if rect is None:
            self._rect = (self.min_x, self.min_y,
                          self.max_x - self.min_x,
                          self.max_y - self.min_y)

    @property
    def rect(self):
        return self._rect

    @rect.setter
    def rect(self, new_rect):
        self._rect = new_rect
        # figure transform for display rect
        self.translate_x, self.translate_y, rw, rh = new_rect
        self.scale_x = abs(rw / float(self.max_x - self.min_x))
        self.scale_y = abs(rh / float(self.max_y - self.min_y))

    def draw(self):
        glPushMatrix()
        if self._rect is not None:
            glScalef(self.scale_x, self.scale_y, 1)
            glTranslatef(self.translate_x - self.min_x, self.translate_x - self.min_y, 0)
        self.batch.draw()
        glPopMatrix()


window = pyglet.window.Window(width=600, height=300, resizable=True)

dirname = os.path.dirname(__file__)
svg = SVG(os.path.join(dirname, 'hello_world.svg'), rect=(0, 0, 600, 300))


@window.event
def on_draw():
    window.clear()
    svg.draw()


@window.event
def on_resize(w, h):
    print("Resized")
    svg.rect = svg.rect[:2] + (w, h)


pyglet.app.run()



from __future__ import annotations

from ctypes import (
    CFUNCTYPE,
    POINTER,
    Structure,
    Union,
    byref,
    c_byte,
    c_char,
    c_char_p,
    c_int,
    c_int16,
    c_int32,
    c_int64,
    c_long,
    c_short,
    c_size_t,
    c_ubyte,
    c_uint,
    c_uint16,
    c_uint32,
    c_uint64,
    c_ulong,
    c_ushort,
    c_void_p,
)
from typing import Any, Callable, Iterable, Sequence

import pyglet.lib

from .base import FontException

_libfreetype = pyglet.lib.load_library("freetype")

_font_data = {}


def _get_function(name: str, argtypes: Iterable[Any], rtype: Any) -> Callable:
    try:
        func = getattr(_libfreetype, name)
        func.argtypes = argtypes
        func.restype = rtype
        return func
    except AttributeError as e:
        raise ImportError(e)  # noqa: B904


FT_Byte = c_char
FT_Bytes = POINTER(FT_Byte)
FT_Char = c_byte
FT_Int = c_int
FT_UInt = c_uint
FT_Int16 = c_int16
FT_UInt16 = c_uint16
FT_Int32 = c_int32
FT_UInt32 = c_uint32
FT_Int64 = c_int64
FT_UInt64 = c_uint64
FT_Short = c_short
FT_UShort = c_ushort
FT_Long = c_long
FT_ULong = c_ulong
FT_Bool = c_char
FT_Offset = c_size_t
# FT_PtrDist = ?
FT_String = c_char
FT_String_Ptr = c_char_p
FT_Tag = FT_UInt32
FT_Error = c_int
FT_Fixed = c_long
FT_Pointer = c_void_p
FT_Pos = c_long

class TT_OS2(Structure):
    _fields_ = [
        ("version",              c_ushort),
        ("xAvgCharWidth",        c_short),
        ("usWeightClass",        c_ushort),  # This is the key field
        ("usWidthClass",         c_ushort),
        ("fsType",               c_short),
        ("ySubscriptXSize",      c_short),
        ("ySubscriptYSize",      c_short),
        ("ySubscriptXOffset",    c_short),
        ("ySubscriptYOffset",    c_short),
        ("ySuperscriptXSize",    c_short),
        ("ySuperscriptYSize",    c_short),
        ("ySuperscriptXOffset",  c_short),
        ("ySuperscriptYOffset",  c_short),
        ("yStrikeoutSize",       c_short),
        ("yStrikeoutPosition",   c_short),
        ("sFamilyClass",         c_short),
        ("panose",               c_byte * 10),
        ("ulUnicodeRange1",      c_uint),
        ("ulUnicodeRange2",      c_uint),
        ("ulUnicodeRange3",      c_uint),
        ("ulUnicodeRange4",      c_uint),
        ("achVendID",            c_char * 4),
        ("fsSelection",          c_ushort),
        ("usFirstCharIndex",     c_ushort),
        ("usLastCharIndex",      c_ushort),
        ("sTypoAscender",        c_short),
        ("sTypoDescender",       c_short),
        ("sTypoLineGap",         c_short),
        ("usWinAscent",          c_ushort),
        ("usWinDescent",         c_ushort),
        ("ulCodePageRange1",     c_uint),
        ("ulCodePageRange2",     c_uint),
    ]


class FT_Vector(Structure):
    _fields_ = [
        ("x", FT_Pos),
        ("y", FT_Pos),
    ]


class FT_BBox(Structure):
    _fields_ = [
        ("xMin", FT_Pos),
        ("yMin", FT_Pos),
        ("xMax", FT_Pos),
        ("yMax", FT_Pos),
    ]


class FT_Matrix(Structure):
    _fields_ = [
        ("xx", FT_Fixed),
        ("xy", FT_Fixed),
        ("yx", FT_Fixed),
        ("yy", FT_Fixed),
    ]


FT_FWord = c_short
FT_UFWord = c_ushort
FT_F2Dot14 = c_short


class FT_UnitVector(Structure):
    _fields_ = [
        ("x", FT_F2Dot14),
        ("y", FT_F2Dot14),
    ]


FT_F26Dot6 = c_long


class FT_Data(Structure):
    _fields_ = [
        ("pointer", POINTER(FT_Byte)),
        ("length", FT_Int),
    ]


FT_Generic_Finalizer = CFUNCTYPE(None, (c_void_p))


class FT_Generic(Structure):
    _fields_ = [
        ("data", c_void_p),
        ("finalizer", FT_Generic_Finalizer),
    ]


class FT_Bitmap(Structure):
    _fields_ = [
        ("rows", c_uint),
        ("width", c_uint),
        ("pitch", c_int),
        ("buffer", POINTER(c_ubyte)),
        ("num_grays", c_short),
        ("pixel_mode", c_ubyte),
        ("palette_mode", c_ubyte),
        ("palette", c_void_p),
    ]


FT_PIXEL_MODE_NONE = 0
FT_PIXEL_MODE_MONO = 1
FT_PIXEL_MODE_GRAY = 2
FT_PIXEL_MODE_GRAY2 = 3
FT_PIXEL_MODE_GRAY4 = 4
FT_PIXEL_MODE_LCD = 5
FT_PIXEL_MODE_LCD_V = 6
FT_PIXEL_MODE_BGRA = 7

FT_KERNING_DEFAULT = 0
FT_KERNING_UNFITTED = 1
FT_KERNING_UNSCALED = 2


class FT_LibraryRec(Structure):
    _fields_ = [
        ("dummy", c_int),
    ]

    def __del__(self) -> None:
        global _library  # noqa: PLW0603
        try:
            print("FT_LibraryRec.__del__") 
            FT_Done_FreeType(byref(self))
            _library = None
        except:  # noqa: S110, E722
            pass


FT_Library = POINTER(FT_LibraryRec)


class FT_Bitmap_Size(Structure):
    _fields_ = [
        ("height", c_ushort),
        ("width", c_ushort),
        ("size", c_long),
        ("x_ppem", c_long),
        ("y_ppem", c_long),
    ]


class FT_Glyph_Metrics(Structure):
    _fields_ = [
        ("width", FT_Pos),
        ("height", FT_Pos),

        ("horiBearingX", FT_Pos),
        ("horiBearingY", FT_Pos),
        ("horiAdvance", FT_Pos),

        ("vertBearingX", FT_Pos),
        ("vertBearingY", FT_Pos),
        ("vertAdvance", FT_Pos),
    ]

    def dump(self) -> None:
        for (name, _) in self._fields_:
            print("FT_Glyph_Metrics", name, repr(getattr(self, name))) 


FT_Glyph_Format = c_ulong


def FT_IMAGE_TAG(tag: str) -> int:  # noqa: N802
    return (ord(tag[0]) << 24) | (ord(tag[1]) << 16) | (ord(tag[2]) << 8) | ord(tag[3])


FT_GLYPH_FORMAT_NONE = 0
FT_GLYPH_FORMAT_COMPOSITE = FT_IMAGE_TAG("comp")
FT_GLYPH_FORMAT_BITMAP = FT_IMAGE_TAG("bits")
FT_GLYPH_FORMAT_OUTLINE = FT_IMAGE_TAG("outl")
FT_GLYPH_FORMAT_PLOTTER = FT_IMAGE_TAG("plot")


class FT_Outline(Structure):
    _fields_ = [
        ("n_contours", c_short),  # number of contours in glyph
        ("n_points", c_short),  # number of points in the glyph
        ("points", POINTER(FT_Vector)),  # the outline's points
        ("tags", c_char_p),  # the points flags
        ("contours", POINTER(c_short)),  # the contour end points
        ("flags", c_int),  # outline masks
    ]


FT_SubGlyph = c_void_p


class FT_GlyphSlotRec(Structure):
    _fields_ = [
        ("library", FT_Library),
        ("face", c_void_p),
        ("next", c_void_p),
        ("reserved", FT_UInt),
        ("generic", FT_Generic),

        ("metrics", FT_Glyph_Metrics),
        ("linearHoriAdvance", FT_Fixed),
        ("linearVertAdvance", FT_Fixed),
        ("advance", FT_Vector),

        ("format", FT_Glyph_Format),

        ("bitmap", FT_Bitmap),
        ("bitmap_left", FT_Int),
        ("bitmap_top", FT_Int),

        ("outline", FT_Outline),
        ("num_subglyphs", FT_UInt),
        ("subglyphs", FT_SubGlyph),

        ("control_data", c_void_p),
        ("control_len", c_long),

        ("lsb_delta", FT_Pos),
        ("rsb_delta", FT_Pos),

        ("other", c_void_p),

        ("internal", c_void_p),
    ]


FT_GlyphSlot = POINTER(FT_GlyphSlotRec)


class FT_Size_Metrics(Structure):
    _fields_ = [
        ("x_ppem", FT_UShort),  # horizontal pixels per EM
        ("y_ppem", FT_UShort),  # vertical pixels per EM

        ("x_scale", FT_Fixed),  # two scales used to convert font units
        ("y_scale", FT_Fixed),  # to 26.6 frac. pixel coordinates

        ("ascender", FT_Pos),  # ascender in 26.6 frac. pixels
        ("descender", FT_Pos),  # descender in 26.6 frac. pixels
        ("height", FT_Pos),  # text height in 26.6 frac. pixels
        ("max_advance", FT_Pos),  # max horizontal advance, in 26.6 pixels
    ]


class FT_SizeRec(Structure):
    _fields_ = [
        ("face", c_void_p),
        ("generic", FT_Generic),
        ("metrics", FT_Size_Metrics),
        ("internal", c_void_p),
    ]


class FT_StreamDesc(Union):
    _fields_ = [
        ("pointer", c_void_p),
        ("value", c_long)
    ]


FT_Stream_IoFunc = CFUNCTYPE(c_ulong, c_void_p, c_ulong, c_void_p, c_ulong)
FT_Stream_CloseFunc = CFUNCTYPE(None, c_void_p)


class FT_StreamRec(Structure):
    _fields_ = [
        ("base", POINTER(c_ubyte)),  # Pointer to raw font data
        ("size", c_ulong),  # Size of the stream
        ("pos", c_ulong),  # Current read position

        ("descriptor", FT_StreamDesc),  # File descriptor or user data
        ("pathname", FT_StreamDesc),  # Optional pathname

        ("read", FT_Stream_IoFunc),  # Read function
        ("close", FT_Stream_CloseFunc),  # Close function

        ("memory", c_void_p),  # FT_Memory pointer (memory management)
        ("cursor", POINTER(c_ubyte)),  # Current stream position
        ("limit", POINTER(c_ubyte)),  # End of stream data
    ]

FT_Stream = FT_StreamRec
FT_Size = POINTER(FT_SizeRec)


class FT_FaceRec(Structure):
    _fields_ = [
        ("num_faces", FT_Long),
        ("face_index", FT_Long),

        ("face_flags", FT_Long),
        ("style_flags", FT_Long),

        ("num_glyphs", FT_Long),

        ("family_name", FT_String_Ptr),
        ("style_name", FT_String_Ptr),

        ("num_fixed_sizes", FT_Int),
        ("available_sizes", POINTER(FT_Bitmap_Size)),

        ("num_charmaps", FT_Int),
        ("charmaps", c_void_p),

        ("generic", FT_Generic),

        ("bbox", FT_BBox),

        ("units_per_EM", FT_UShort),
        ("ascender", FT_Short),
        ("descender", FT_Short),
        ("height", FT_Short),

        ("max_advance_width", FT_Short),
        ("max_advance_height", FT_Short),

        ("underline_position", FT_Short),
        ("underline_thickness", FT_Short),

        ("glyph", FT_GlyphSlot),
        ("size", FT_Size),
        ("charmap", c_void_p),

        # Internal
        ("driver", c_void_p),
        ("memory", c_void_p),
        ("stream", POINTER(FT_Stream)),

        ("sizes_list", c_void_p),

        ("autohint", FT_Generic),
        ("extensions", c_void_p),
        ("internal", c_void_p),
    ]

    def dump(self) -> None:
        for (name, _) in self._fields_:
            print("FT_FaceRec", name, repr(getattr(self, name)))


FT_Face = POINTER(FT_FaceRec)

# face_flags values
FT_FACE_FLAG_SCALABLE          = 1 << 0   # Vector-based (TTF/OTF
FT_FACE_FLAG_FIXED_SIZES       = 1 << 1   # Has pre-rendered bitmap sizes
FT_FACE_FLAG_FIXED_WIDTH       = 1 << 2   # Monospaced font
FT_FACE_FLAG_SFNT              = 1 << 3   # SFNT-based (TrueType/OpenType/WOFF
FT_FACE_FLAG_HORIZONTAL        = 1 << 4   # Has horizontal glyph metrics
FT_FACE_FLAG_VERTICAL          = 1 << 5   # Has vertical glyph metrics
FT_FACE_FLAG_KERNING           = 1 << 6   # Supports kerning
FT_FACE_FLAG_FAST_GLYPHS       = 1 << 7   # Deprecated (No effect
FT_FACE_FLAG_MULTIPLE_MASTERS  = 1 << 8   # Supports variable fonts (GX/AAT
FT_FACE_FLAG_GLYPH_NAMES       = 1 << 9   # Has glyph names
FT_FACE_FLAG_EXTERNAL_STREAM   = 1 << 10  # Uses external stream (not file-based
FT_FACE_FLAG_HINTER            = 1 << 11  # Font has its own hinter
FT_FACE_FLAG_CID_KEYED         = 1 << 12  # CID-keyed PostScript font
FT_FACE_FLAG_TRICKY            = 1 << 13  # Requires special TrueType bytecode handling
FT_FACE_FLAG_COLOR             = 1 << 14  # Color emoji font
FT_FACE_FLAG_VARIATION         = 1 << 15  # OpenType Variable Fonts
FT_FACE_FLAG_SVG               = 1 << 16  # SVG-based color glyphs
FT_FACE_FLAG_SBIX              = 1 << 17  # sbix color glyphs
FT_FACE_FLAG_SBIX_OVERLAY      = 1 << 18  # sbix overlay support

FT_STYLE_FLAG_ITALIC = 1
FT_STYLE_FLAG_BOLD = 2

(FT_RENDER_MODE_NORMAL,
 FT_RENDER_MODE_LIGHT,
 FT_RENDER_MODE_MONO,
 FT_RENDER_MODE_LCD,
 FT_RENDER_MODE_LCD_V) = range(5)


def FT_LOAD_TARGET_(x: int) -> int:  # noqa: N802
    return (x & 15) << 16


FT_LOAD_TARGET_NORMAL = FT_LOAD_TARGET_(FT_RENDER_MODE_NORMAL)
FT_LOAD_TARGET_LIGHT = FT_LOAD_TARGET_(FT_RENDER_MODE_LIGHT)
FT_LOAD_TARGET_MONO = FT_LOAD_TARGET_(FT_RENDER_MODE_MONO)
FT_LOAD_TARGET_LCD = FT_LOAD_TARGET_(FT_RENDER_MODE_LCD)
FT_LOAD_TARGET_LCD_V = FT_LOAD_TARGET_(FT_RENDER_MODE_LCD_V)


def f16p16_to_float(value: float) -> float:
    return float(value) / (1 << 16)


def float_to_f16p16(value: float) -> float:
    return int(value * (1 << 16))


def f26p6_to_float(value: float) -> float:
    return float(value) / (1 << 6)


def float_to_f26p6(value: float) -> float:
    return int(value * (1 << 6))



_ft_errors = {
    0x00: "no error",
    0x01: "cannot open resource",
    0x02: "unknown file format",
    0x03: "broken file",
    0x04: "invalid FreeType version",
    0x05: "module version is too low",
    0x06: "invalid argument",
    0x07: "unimplemented feature",
    0x08: "broken table",
    0x09: "broken offset within table",
    0x10: "invalid glyph index",
    0x11: "invalid character code",
    0x12: "unsupported glyph image format",
    0x13: "cannot render this glyph format",
    0x14: "invalid outline",
    0x15: "invalid composite glyph",
    0x16: "too many hints",
    0x17: "invalid pixel size",
    0x20: "invalid object handle",
    0x21: "invalid library handle",
    0x22: "invalid module handle",
    0x23: "invalid face handle",
    0x24: "invalid size handle",
    0x25: "invalid glyph slot handle",
    0x26: "invalid charmap handle",
    0x27: "invalid cache manager handle",
    0x28: "invalid stream handle",
    0x30: "too many modules",
    0x31: "too many extensions",
    0x40: "out of memory",
    0x41: "unlisted object",
    0x51: "cannot open stream",
    0x52: "invalid stream seek",
    0x53: "invalid stream skip",
    0x54: "invalid stream read",
    0x55: "invalid stream operation",
    0x56: "invalid frame operation",
    0x57: "nested frame access",
    0x58: "invalid frame read",
    0x60: "raster uninitialized",
    0x61: "raster corrupted",
    0x62: "raster overflow",
    0x63: "negative height while rastering",
    0x70: "too many registered caches",
    0x80: "invalid opcode",
    0x81: "too few arguments",
    0x82: "stack overflow",
    0x83: "code overflow",
    0x84: "bad argument",
    0x85: "division by zero",
    0x86: "invalid reference",
    0x87: "found debug opcode",
    0x88: "found ENDF opcode in execution stream",
    0x89: "nested DEFS",
    0x8A: "invalid code range",
    0x8B: "execution context too long",
    0x8C: "too many function definitions",
    0x8D: "too many instruction definitions",
    0x8E: "SFNT font table missing",
    0x8F: "horizontal header (hhea, table missing",
    0x90: "locations (loca, table missing",
    0x91: "name table missing",
    0x92: "character map (cmap, table missing",
    0x93: "horizontal metrics (hmtx, table missing",
    0x94: "PostScript (post, table missing",
    0x95: "invalid horizontal metrics",
    0x96: "invalid character map (cmap, format",
    0x97: "invalid ppem value",
    0x98: "invalid vertical metrics",
    0x99: "could not find context",
    0x9A: "invalid PostScript (post, table format",
    0x9B: "invalid PostScript (post, table",
    0xA0: "opcode syntax error",
    0xA1: "argument stack underflow",
    0xA2: "ignore",
    0xB0: "`STARTFONT' field missing",
    0xB1: "`FONT' field missing",
    0xB2: "`SIZE' field missing",
    0xB3: "`CHARS' field missing",
    0xB4: "`STARTCHAR' field missing",
    0xB5: "`ENCODING' field missing",
    0xB6: "`BBX' field missing",
    0xB7: "`BBX' too big",
}


class FreeTypeError(FontException):
    def __init__(self, error_code, message):
        self.errcode = error_code
        self.message = message


def _errcheck(result: Any, func: Callable, arguments: Sequence) -> Any:
    if result != 0:
        raise FreeTypeError(result, _ft_errors.get(result, "Unknown."))

def _get_function_with_error_handling(name: str, argtypes: Iterable[Any], rtype: Any) -> Callable:
    func = _get_function(name, argtypes, rtype)
    func.errcheck = _errcheck
    func.__name__ = name
    return func


FT_LOAD_DEFAULT = 0x0  # Use default font loading behavior
FT_LOAD_NO_SCALE = (1 << 0)  # Disable scaling; use unscaled native units
FT_LOAD_NO_HINTING = (1 << 1)  # Disable font hinting (may cause blurriness at small sizes)
FT_LOAD_RENDER = (1 << 2)  # Automatically render the glyph after loading
FT_LOAD_NO_BITMAP = (1 << 3)  # Ignore bitmap strikes, force vector rendering
FT_LOAD_VERTICAL_LAYOUT = (1 << 4)  # Load glyph with vertical metrics if available
FT_LOAD_FORCE_AUTOHINT = (1 << 5)  # Force FreeType’s auto-hinter, ignoring native hints
FT_LOAD_CROP_BITMAP = (1 << 6)  # Crop bitmaps to bounding box (removes empty padding)
FT_LOAD_PEDANTIC = (1 << 7)  # Enable strict font parsing (useful for debugging malformed fonts)
FT_LOAD_IGNORE_GLOBAL_ADVANCE_WIDTH = (1 << 9)  # Ignore font-wide advance width; use glyph-specific advance
FT_LOAD_NO_RECURSE = (1 << 10)  # Load only composite glyphs, do not load sub-glyphs recursively
FT_LOAD_IGNORE_TRANSFORM = (1 << 11)  # Disable any transformations with FT_Set_Transform()
FT_LOAD_MONOCHROME = (1 << 12)  # Render glyphs in monochrome (1-bit, no anti-aliasing)
FT_LOAD_LINEAR_DESIGN = (1 << 13)  # Use linear design metrics instead of transformed metrics
FT_LOAD_SBITS_ONLY = (1 << 14)  # Load only embedded bitmap strikes (ignore vector outlines)
FT_LOAD_NO_AUTOHINT = (1 << 15)  # Disable FreeType’s auto-hinter
FT_LOAD_COLOR = (1 << 20)  # Load color emoji glyphs (e.g., sbix, CBDT/CBLC, SVG fonts)
FT_LOAD_COMPUTE_METRICS = (1 << 21)  # Compute glyph metrics even if loading only a bitmap
FT_LOAD_BITMAP_METRICS_ONLY = (1 << 22)  # Compute bitmap glyph metrics but do not load the glyph itself
FT_LOAD_NO_SVG = (1 << 24)  # Disable loading of SVG glyphs (for fonts that contain both vector and SVG glyphs)

FT_Sfnt_Tag = c_int
class FT_Sfnt_Tags:
    """SFNT table tags for use with FT_Get_Sfnt_Table."""
    HEAD  = 0
    MAXP  = 1
    OS2   = 2
    HHEA  = 3
    VHEA  = 4
    POST  = 5
    NAME  = 6
    CFF   = 7
    GASP  = 8
    PCLT  = 9


FT_Init_FreeType = _get_function_with_error_handling("FT_Init_FreeType",
                                                     [POINTER(FT_Library)], FT_Error)
FT_Done_FreeType = _get_function_with_error_handling("FT_Done_FreeType",
                                                     [FT_Library], FT_Error)

FT_New_Face = _get_function_with_error_handling("FT_New_Face",
                                                [FT_Library, c_char_p, FT_Long, POINTER(FT_Face)], FT_Error)
FT_Done_Face = _get_function_with_error_handling("FT_Done_Face",
                                                 [FT_Face], FT_Error)
FT_Reference_Face = _get_function_with_error_handling("FT_Reference_Face",
                                                      [FT_Face], FT_Error)
FT_New_Memory_Face = _get_function_with_error_handling("FT_New_Memory_Face",
                                                       [FT_Library, POINTER(FT_Byte), FT_Long, FT_Long,
                                                        POINTER(FT_Face)], FT_Error)

FT_Set_Char_Size = _get_function_with_error_handling("FT_Set_Char_Size",
                                                     [FT_Face, FT_F26Dot6, FT_F26Dot6, FT_UInt, FT_UInt], FT_Error)
FT_Set_Pixel_Sizes = _get_function_with_error_handling("FT_Set_Pixel_Sizes",
                                                       [FT_Face, FT_UInt, FT_UInt], FT_Error)
FT_Load_Glyph = _get_function_with_error_handling("FT_Load_Glyph",
                                                  [FT_Face, FT_UInt, FT_Int32], FT_Error)
FT_Get_Char_Index = _get_function("FT_Get_Char_Index",
                                                      [FT_Face, FT_ULong], FT_UInt)
FT_Load_Char = _get_function_with_error_handling("FT_Load_Char",
                                                 [FT_Face, FT_ULong, FT_Int32], FT_Error)
FT_Get_Kerning = _get_function_with_error_handling("FT_Get_Kerning",
                                                   [FT_Face, FT_UInt, FT_UInt, FT_UInt, POINTER(FT_Vector)], FT_Error)

FT_Get_Sfnt_Table = _get_function("FT_Get_Sfnt_Table", [FT_Face, FT_Sfnt_Tag], c_void_p)

try:
    FT_Select_Size = _get_function_with_error_handling("FT_Select_Size", [FT_Face, FT_Int], FT_Error)
except ImportError:
    FT_Select_Size = None

# SFNT interface

class FT_SfntName(Structure):
    _fields_ = [
        ("platform_id", FT_UShort),
        ("encoding_id", FT_UShort),
        ("language_id", FT_UShort),
        ("name_id", FT_UShort),
        ("string", POINTER(FT_Byte)),
        ("string_len", FT_UInt),
    ]


FT_Get_Sfnt_Name_Count = _get_function("FT_Get_Sfnt_Name_Count",
                                       [FT_Face], FT_UInt)
FT_Get_Sfnt_Name = _get_function_with_error_handling("FT_Get_Sfnt_Name",
                                                     [FT_Face, FT_UInt, POINTER(FT_SfntName)], FT_Error)

TT_PLATFORM_MICROSOFT = 3
TT_MS_ID_UNICODE_CS = 1
TT_NAME_ID_COPYRIGHT = 0
TT_NAME_ID_FONT_FAMILY = 1
TT_NAME_ID_FONT_SUBFAMILY = 2
TT_NAME_ID_UNIQUE_ID = 3
TT_NAME_ID_FULL_NAME = 4
TT_NAME_ID_VERSION_STRING = 5
TT_NAME_ID_PS_NAME = 6
TT_NAME_ID_TRADEMARK = 7
TT_NAME_ID_MANUFACTURER = 8
TT_NAME_ID_DESIGNER = 9
TT_NAME_ID_DESCRIPTION = 10
TT_NAME_ID_VENDOR_URL = 11
TT_NAME_ID_DESIGNER_URL = 12
TT_NAME_ID_LICENSE = 13
TT_NAME_ID_LICENSE_URL = 14
TT_NAME_ID_PREFERRED_FAMILY = 16
TT_NAME_ID_PREFERRED_SUBFAMILY = 17
TT_NAME_ID_MAC_FULL_NAME = 18
TT_NAME_ID_CID_FINDFONT_NAME = 20

_library = None


def ft_get_library() -> FT_Library:
    global _library  # noqa: PLW0603
    if not _library:
        _library = FT_Library()
        error = FT_Init_FreeType(byref(_library))
        if error:
            msg = "an error occurred during library initialization"
            raise FontException(msg, error)
    return _library
